from typing import Optional, Sequence
from besser.BUML.metamodel.gui.graphical_ui import ViewComponent
from besser.BUML.metamodel.gui.style import Alignment
from besser.BUML.metamodel.gui.binding import DataBinding
from besser.BUML.metamodel.gui.style import Styling
from besser.BUML.metamodel.structural import Property

class AgentComponent(ViewComponent):
    """Represents a BESSER Agent component in the GUI.
    
    Args:
        name (str): The name of the agent component.
        agent_name (str | None): The name of the agent diagram to reference.
        agent_title (str | None): Display title for the agent component.
    """

    def __init__(
        self,
        name: str,
        agent_name: Optional[str] = None,
        agent_title: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self.agent_name = agent_name
        self.agent_title = agent_title


    @property
    def agent_name(self) -> Optional[str]:
        """Optional[str]: Name of the agent diagram to reference."""
        return self._agent_name

    @agent_name.setter
    def agent_name(self, value: Optional[str]):
        self._agent_name = value

    @property
    def agent_title(self) -> Optional[str]:
        """Optional[str]: Display title for the agent component."""
        return self._agent_title

    @agent_title.setter
    def agent_title(self, value: Optional[str]):
        self._agent_title = value

class Series(ViewComponent):
    """Represents a data series in a chart.

    Args:
        name (str): The name of the series.
        label (str | None): Label for the series.
        data_binding (DataBinding): The data binding for the series.
        styling (Styling): The styling for the series.
    """
    def __init__(self, name: str, label: Optional[str], data_binding: DataBinding, styling: Styling):
        super().__init__(name, data_binding=data_binding, styling=styling)
        self.label = label

    def __repr__(self):
        return f"Series(name={self.name}, label={self.label}, data_binding={self.data_binding}, styling={self.styling})"

class Chart(ViewComponent):
    """Represents a chart component in the dashboard.

    Args:
        name (str): The name of the chart.
        title (str | None): Optional display title of the chart.
        primary_color (str | None): Optional primary color used by the chart.
        series (list[Series]): List of data series in the chart.
    """

    def __init__(
        self,
        name: str,
        series: Optional[Sequence[Series]] = None,
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self.title = title
        self.primary_color = primary_color
        self.series = series

    @property
    def title(self) -> Optional[str]:
        """Optional[str]: Display title for the chart."""
        return self._title

    @title.setter
    def title(self, value: Optional[str]):
        self._title = value

    @property
    def primary_color(self) -> Optional[str]:
        """Optional[str]: Main color used by the chart."""
        return self._primary_color

    @primary_color.setter
    def primary_color(self, value: Optional[str]):
        self._primary_color = value
    
    @property
    def series(self) -> list[Series]:
        """list[Series]: List of data series in the chart."""
        return self._series

    @series.setter
    def series(self, value: Optional[Sequence[Series]]):
        if value is None:
            self._series = []
        else:
            self._series = list(value)

class LineChart(Chart):
    """Represents a line chart component in the dashboard.

    Args:
        name (str): The name of the line chart.
        line_width (int): The width of the line in the chart.
        show_grid (bool): Whether to show the grid in the chart.
        show_legend (bool): Whether to show the legend in the chart.
        show_tooltip (bool): Whether to show tooltips.
        curve_type (str): The curve type ('linear', 'monotone', 'step').
        animate (bool): Whether to animate the chart.
        legend_position (str): Position of the legend ('top', 'right', 'bottom', 'left').
        grid_color (str): Color of the grid lines.
        dot_size (int): Size of the data point dots.

    Attributes:
        name (str): The name of the line chart.
        line_width (int): The width of the line in the chart.
        show_grid (bool): Whether to show the grid.
        show_legend (bool): Whether to show the legend.
        show_tooltip (bool): Whether to show tooltips.
        curve_type (str): The curve type.
        animate (bool): Whether to animate the chart.
        legend_position (str): Position of the legend.
        grid_color (str): Color of the grid lines.
        dot_size (int): Size of the data point dots.
    """

    def __init__(
        self,
        name: str,
        line_width: int = 2,
        show_grid: bool = True,
        show_legend: bool = True,
        show_tooltip: bool = True,
        curve_type: str = "monotone",
        animate: bool = True,
        legend_position: str = "top",
        grid_color: str = "#e0e0e0",
        dot_size: int = 5,
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, title=title, primary_color=primary_color, **kwargs)
        self.line_width = line_width
        self.show_grid = show_grid
        self.show_legend = show_legend
        self.show_tooltip = show_tooltip
        self.curve_type = curve_type
        self.animate = animate
        self.legend_position = legend_position
        self.grid_color = grid_color
        self.dot_size = dot_size

    @property
    def line_width(self) -> int:
        """Property: Get the line width of the line chart."""
        return self._line_width

    @line_width.setter
    def line_width(self, value: int):
        """Property: Set the line width of the line chart."""
        self._line_width = value

    @property
    def show_grid(self) -> bool:
        """Property: Get whether to show grid."""
        return self._show_grid

    @show_grid.setter
    def show_grid(self, value: bool):
        """Property: Set whether to show grid."""
        self._show_grid = value

    @property
    def show_legend(self) -> bool:
        """Property: Get whether to show legend."""
        return self._show_legend

    @show_legend.setter
    def show_legend(self, value: bool):
        """Property: Set whether to show legend."""
        self._show_legend = value

    @property
    def show_tooltip(self) -> bool:
        """Property: Get whether to show tooltip."""
        return self._show_tooltip

    @show_tooltip.setter
    def show_tooltip(self, value: bool):
        """Property: Set whether to show tooltip."""
        self._show_tooltip = value

    @property
    def curve_type(self) -> str:
        """Property: Get the curve type."""
        return self._curve_type

    @curve_type.setter
    def curve_type(self, value: str):
        """Property: Set the curve type."""
        self._curve_type = value

    @property
    def animate(self) -> bool:
        """Property: Get whether to animate."""
        return self._animate

    @animate.setter
    def animate(self, value: bool):
        """Property: Set whether to animate."""
        self._animate = value

    @property
    def legend_position(self) -> str:
        """Property: Get legend position."""
        return self._legend_position

    @legend_position.setter
    def legend_position(self, value: str):
        """Property: Set legend position."""
        self._legend_position = value

    @property
    def grid_color(self) -> str:
        """Property: Get grid color."""
        return self._grid_color

    @grid_color.setter
    def grid_color(self, value: str):
        """Property: Set grid color."""
        self._grid_color = value

    @property
    def dot_size(self) -> int:
        """Property: Get dot size."""
        return self._dot_size

    @dot_size.setter
    def dot_size(self, value: int):
        """Property: Set dot size."""
        self._dot_size = value

    def __repr__(self):
        return (
            f"LineChart(name={self.name}, line_width={self.line_width}, "
            f"show_grid={self.show_grid}, show_legend={self.show_legend}, "
            f"curve_type={self.curve_type})"
        )

class BarChart(Chart):
    """Represents a bar chart component in the dashboard.

    Args:
        name (str): The name of the bar chart.
        bar_width (int): The width of the bars in the chart.
        orientation (str): Orientation of bars ('vertical' or 'horizontal').
        show_grid (bool): Whether to show the grid.
        show_legend (bool): Whether to show the legend.
        show_tooltip (bool): Whether to show tooltips.
        stacked (bool): Whether to stack bars.
        animate (bool): Whether to animate the chart.
        legend_position (str): Position of the legend.
        grid_color (str): Color of the grid lines.
        bar_gap (int): Gap between bars.

    Attributes:
        name (str): The name of the bar chart.
        bar_width (int): The width of the bars in the chart.
        orientation (str): Orientation of bars.
        show_grid (bool): Whether to show the grid.
        show_legend (bool): Whether to show the legend.
        show_tooltip (bool): Whether to show tooltips.
        stacked (bool): Whether to stack bars.
        animate (bool): Whether to animate.
        legend_position (str): Position of the legend.
        grid_color (str): Color of the grid.
        bar_gap (int): Gap between bars.
    """

    def __init__(
        self,
        name: str,
        bar_width: int = 30,
        orientation: str = "vertical",
        show_grid: bool = True,
        show_legend: bool = True,
        show_tooltip: bool = True,
        stacked: bool = False,
        animate: bool = True,
        legend_position: str = "top",
        grid_color: str = "#e0e0e0",
        bar_gap: int = 4,
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, title=title, primary_color=primary_color, **kwargs)
        self.bar_width = bar_width
        self.orientation = orientation
        self.show_grid = show_grid
        self.show_legend = show_legend
        self.show_tooltip = show_tooltip
        self.stacked = stacked
        self.animate = animate
        self.legend_position = legend_position
        self.grid_color = grid_color
        self.bar_gap = bar_gap

    @property
    def bar_width(self) -> int:
        """Property: Get the bar width of the bar chart."""
        return self._bar_width

    @bar_width.setter
    def bar_width(self, value: int):
        """Property: Set the bar width of the bar chart."""
        self._bar_width = value

    @property
    def orientation(self) -> str:
        """Property: Get orientation."""
        return self._orientation

    @orientation.setter
    def orientation(self, value: str):
        """Property: Set orientation."""
        self._orientation = value

    @property
    def show_grid(self) -> bool:
        """Property: Get whether to show grid."""
        return self._show_grid

    @show_grid.setter
    def show_grid(self, value: bool):
        """Property: Set whether to show grid."""
        self._show_grid = value

    @property
    def show_legend(self) -> bool:
        """Property: Get whether to show legend."""
        return self._show_legend

    @show_legend.setter
    def show_legend(self, value: bool):
        """Property: Set whether to show legend."""
        self._show_legend = value

    @property
    def show_tooltip(self) -> bool:
        """Property: Get whether to show tooltip."""
        return self._show_tooltip

    @show_tooltip.setter
    def show_tooltip(self, value: bool):
        """Property: Set whether to show tooltip."""
        self._show_tooltip = value

    @property
    def stacked(self) -> bool:
        """Property: Get whether bars are stacked."""
        return self._stacked

    @stacked.setter
    def stacked(self, value: bool):
        """Property: Set whether bars are stacked."""
        self._stacked = value

    @property
    def animate(self) -> bool:
        """Property: Get whether to animate."""
        return self._animate

    @animate.setter
    def animate(self, value: bool):
        """Property: Set whether to animate."""
        self._animate = value

    @property
    def legend_position(self) -> str:
        """Property: Get legend position."""
        return self._legend_position

    @legend_position.setter
    def legend_position(self, value: str):
        """Property: Set legend position."""
        self._legend_position = value

    @property
    def grid_color(self) -> str:
        """Property: Get grid color."""
        return self._grid_color

    @grid_color.setter
    def grid_color(self, value: str):
        """Property: Set grid color."""
        self._grid_color = value

    @property
    def bar_gap(self) -> int:
        """Property: Get bar gap."""
        return self._bar_gap

    @bar_gap.setter
    def bar_gap(self, value: int):
        """Property: Set bar gap."""
        self._bar_gap = value

    def __repr__(self):
        return (
            f"BarChart(name={self.name}, bar_width={self.bar_width}, "
            f"orientation={self.orientation}, stacked={self.stacked})"
        )

class PieChart(Chart):
    """Represents a pie chart component in the dashboard.

    Args:
        name (str): The name of the pie chart.
        show_legend (bool): Whether to show the legend in the pie chart.
        legend_position (Alignment): The position of the legend in the pie chart.
        show_labels (bool): Whether to show labels in the pie chart.
        label_position (Alignment): The position of the labels in the pie chart.
        padding_angle (int): The padding angle between slices in the pie chart.
        inner_radius (int): Inner radius for donut charts (0-100).
        outer_radius (int): Outer radius percentage (0-100).
        start_angle (int): Start angle in degrees.
        end_angle (int): End angle in degrees.

    Attributes:
        name (str): The name of the pie chart.
        show_legend (bool): Whether to show the legend in the pie chart.
        legend_position (Alignment): The position of the legend in the pie chart.
        show_labels (bool): Whether to show labels in the pie chart.
        label_position (Alignment): The position of the labels in the pie chart.
        padding_angle (int): The padding angle between slices in the pie chart.
        inner_radius (int): Inner radius for donut charts.
        outer_radius (int): Outer radius percentage.
        start_angle (int): Start angle in degrees.
        end_angle (int): End angle in degrees.
    """

    def __init__(
        self,
        name: str,
        show_legend: bool = True,
        legend_position: Alignment = Alignment.LEFT,
        show_labels: bool = True,
        label_position: Alignment = Alignment.INSIDE,
        padding_angle: int = 0,
        inner_radius: int = 0,
        outer_radius: int = 80,
        start_angle: int = 0,
        end_angle: int = 360,
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, title=title, primary_color=primary_color, **kwargs)
        self.show_legend: bool = show_legend
        self.legend_position: Alignment = legend_position
        self.show_labels: bool = show_labels
        self.label_position: Alignment = label_position
        self.padding_angle: int = padding_angle
        self.inner_radius: int = inner_radius
        self.outer_radius: int = outer_radius
        self.start_angle: int = start_angle
        self.end_angle: int = end_angle

    @property
    def show_legend(self) -> bool:
        """Property: Get whether to show the legend of the pie chart."""
        return self._show_legend

    @show_legend.setter
    def show_legend(self, value: bool):
        """Property: Set whether to show the legend of the pie chart."""
        self._show_legend = value

    @property
    def legend_position(self) -> Alignment:
        """Property: Get the legend position of the pie chart."""
        return self._legend_position

    @legend_position.setter
    def legend_position(self, value: Alignment):
        """Property: Set the legend position of the pie chart."""
        self._legend_position = value

    @property
    def show_labels(self) -> bool:
        """Property: Get whether to show the labels of the pie chart."""
        return self._show_labels

    @show_labels.setter
    def show_labels(self, value: bool): 
        """Property: Set whether to show the labels of the pie chart."""
        self._show_labels = value

    @property
    def label_position(self) -> Alignment:
        """Property: Get the label position of the pie chart."""
        return self._label_position

    @label_position.setter
    def label_position(self, value: Alignment):
        """Property: Set the label position of the pie chart."""
        self._label_position = value

    @property
    def padding_angle(self) -> int:
        """Property: Get the padding angle of the pie chart."""
        return self._padding_angle

    @padding_angle.setter
    def padding_angle(self, value: int):
        """Property: Set the padding angle of the pie chart."""
        self._padding_angle = value

    @property
    def inner_radius(self) -> int:
        """Property: Get the inner radius."""
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value: int):
        """Property: Set the inner radius."""
        self._inner_radius = value

    @property
    def outer_radius(self) -> int:
        """Property: Get the outer radius."""
        return self._outer_radius

    @outer_radius.setter
    def outer_radius(self, value: int):
        """Property: Set the outer radius."""
        self._outer_radius = value

    @property
    def start_angle(self) -> int:
        """Property: Get the start angle."""
        return self._start_angle

    @start_angle.setter
    def start_angle(self, value: int):
        """Property: Set the start angle."""
        self._start_angle = value

    @property
    def end_angle(self) -> int:
        """Property: Get the end angle."""
        return self._end_angle

    @end_angle.setter
    def end_angle(self, value: int):
        """Property: Set the end angle."""
        self._end_angle = value

    def __repr__(self):
        return (
            f"PieChart(name={self.name}, show_legend={self.show_legend}, "
            f"legend_position={self.legend_position}, show_labels={self.show_labels}, "
            f"label_position={self.label_position}, padding_angle={self.padding_angle})"
        )

class RadarChart(Chart):
    """Represents a radar chart component in the dashboard.

    Args:
        name (str): The name of the radar chart.
        show_grid (bool): Whether to show the grid in the radar chart.
        show_tooltip (bool): Whether to show the tooltip in the radar chart.
        show_radius_axis (bool): Whether to show the radius axis in the radar chart.
        show_legend (bool): Whether to show the legend.
        legend_position (str): Position of the legend.
        dot_size (int): Size of data point dots.
        grid_type (str): Type of grid ('polygon' or 'circle').
        stroke_width (int): Width of the stroke lines.

    Attributes:
        name (str): The name of the radar chart.
        show_grid (bool): Whether to show the grid in the radar chart.
        show_tooltip (bool): Whether to show the tooltip in the radar chart.
        show_radius_axis (bool): Whether to show the radius axis in the radar chart.
        show_legend (bool): Whether to show the legend.
        legend_position (str): Position of the legend.
        dot_size (int): Size of data point dots.
        grid_type (str): Type of grid.
        stroke_width (int): Width of the stroke lines.
    """

    def __init__(
        self,
        name: str,
        show_grid: bool = True,
        show_tooltip: bool = True,
        show_radius_axis: bool = True,
        show_legend: bool = True,
        legend_position: str = "top",
        dot_size: int = 3,
        grid_type: str = "polygon",
        stroke_width: int = 2,
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, title=title, primary_color=primary_color, **kwargs)
        self.show_grid = show_grid
        self.show_tooltip = show_tooltip
        self.show_radius_axis = show_radius_axis
        self.show_legend = show_legend
        self.legend_position = legend_position
        self.dot_size = dot_size
        self.grid_type = grid_type
        self.stroke_width = stroke_width

    @property
    def show_grid(self) -> bool:
        """Property: Get the show_grid option of the radar chart."""
        return self._show_grid

    @show_grid.setter
    def show_grid(self, value: bool):
        """Property: Set the show_grid option of the radar chart."""
        self._show_grid = value

    @property
    def show_tooltip(self) -> bool:
        """Property: Get the show_tooltip option of the radar chart."""
        return self._show_tooltip

    @show_tooltip.setter
    def show_tooltip(self, value: bool):
        """Property: Set the show_tooltip option of the radar chart."""
        self._show_tooltip = value

    @property
    def show_radius_axis(self) -> bool:
        """Property: Get the show_radius_axis option of the radar chart."""
        return self._show_radius_axis

    @show_radius_axis.setter
    def show_radius_axis(self, value: bool):
        """Property: Set the show_radius_axis option of the radar chart."""
        self._show_radius_axis = value

    @property
    def show_legend(self) -> bool:
        """Property: Get whether to show legend."""
        return self._show_legend

    @show_legend.setter
    def show_legend(self, value: bool):
        """Property: Set whether to show legend."""
        self._show_legend = value

    @property
    def legend_position(self) -> str:
        """Property: Get legend position."""
        return self._legend_position

    @legend_position.setter
    def legend_position(self, value: str):
        """Property: Set legend position."""
        self._legend_position = value

    @property
    def dot_size(self) -> int:
        """Property: Get dot size."""
        return self._dot_size

    @dot_size.setter
    def dot_size(self, value: int):
        """Property: Set dot size."""
        self._dot_size = value

    @property
    def grid_type(self) -> str:
        """Property: Get grid type."""
        return self._grid_type

    @grid_type.setter
    def grid_type(self, value: str):
        """Property: Set grid type."""
        self._grid_type = value

    @property
    def stroke_width(self) -> int:
        """Property: Get stroke width."""
        return self._stroke_width

    @stroke_width.setter
    def stroke_width(self, value: int):
        """Property: Set stroke width."""
        self._stroke_width = value

    def __repr__(self):
        return (
            f"RadarChart(name={self.name}, show_grid={self.show_grid}, show_tooltip={self.show_tooltip}, "
            f"show_radius_axis={self.show_radius_axis}, grid_type={self.grid_type})"
        )

class RadialBarChart(Chart):
    """Represents a radial bar chart component in the dashboard.

    Args:
        name (str): The name of the radial bar chart.
        start_angle (int): The start angle of the radial bar chart.
        end_angle (int): The end angle of the radial bar chart.
        inner_radius (int): Inner radius percentage (0-100).
        outer_radius (int): Outer radius percentage (0-100).
        show_legend (bool): Whether to show the legend.
        legend_position (str): Position of the legend.
        show_tooltip (bool): Whether to show tooltips.
        
    Attributes:
        name (str): The name of the radial bar chart.
        start_angle (int): The start angle of the radial bar chart.
        end_angle (int): The end angle of the radial bar chart.
        inner_radius (int): Inner radius percentage.
        outer_radius (int): Outer radius percentage.
        show_legend (bool): Whether to show the legend.
        legend_position (str): Position of the legend.
        show_tooltip (bool): Whether to show tooltips.
    """

    def __init__(
        self,
        name: str,
        start_angle: int = 0,
        end_angle: int = 360,
        inner_radius: int = 30,
        outer_radius: int = 80,
        show_legend: bool = True,
        legend_position: str = "top",
        show_tooltip: bool = True,
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, title=title, primary_color=primary_color, **kwargs)
        self.start_angle = start_angle
        self.end_angle = end_angle
        self.inner_radius = inner_radius
        self.outer_radius = outer_radius
        self.show_legend = show_legend
        self.legend_position = legend_position
        self.show_tooltip = show_tooltip

    @property
    def start_angle(self) -> int:
        """Property: Get the start angle of the radial bar chart."""
        return self._start_angle

    @start_angle.setter
    def start_angle(self, value: int):
        """Property: Set the start angle of the radial bar chart."""
        self._start_angle = value

    @property
    def end_angle(self) -> int:
        """Property: Get the end angle of the radial bar chart."""
        return self._end_angle

    @end_angle.setter
    def end_angle(self, value: int):
        """Property: Set the end angle of the radial bar chart."""
        self._end_angle = value

    @property
    def inner_radius(self) -> int:
        """Property: Get the inner radius."""
        return self._inner_radius

    @inner_radius.setter
    def inner_radius(self, value: int):
        """Property: Set the inner radius."""
        self._inner_radius = value

    @property
    def outer_radius(self) -> int:
        """Property: Get the outer radius."""
        return self._outer_radius

    @outer_radius.setter
    def outer_radius(self, value: int):
        """Property: Set the outer radius."""
        self._outer_radius = value

    @property
    def show_legend(self) -> bool:
        """Property: Get whether to show legend."""
        return self._show_legend

    @show_legend.setter
    def show_legend(self, value: bool):
        """Property: Set whether to show legend."""
        self._show_legend = value

    @property
    def legend_position(self) -> str:
        """Property: Get legend position."""
        return self._legend_position

    @legend_position.setter
    def legend_position(self, value: str):
        """Property: Set legend position."""
        self._legend_position = value

    @property
    def show_tooltip(self) -> bool:
        """Property: Get whether to show tooltip."""
        return self._show_tooltip

    @show_tooltip.setter
    def show_tooltip(self, value: bool):
        """Property: Set whether to show tooltip."""
        self._show_tooltip = value

    def __repr__(self):
        return (
            f"RadialBarChart(name={self.name}, start_angle={self.start_angle}, "
            f"end_angle={self.end_angle}, inner_radius={self.inner_radius})"
        )

class Column:
    """Represents a column in a table.

    Args:
        label (str): The display label of the column.
    
    Attributes:
        label (str): The display label of the column.
    """
    def __init__(self, label: str):
        self.label = label

    @property
    def label(self) -> str:
        """Property: Get the label of the column."""
        return self._label

    @label.setter
    def label(self, value: str):
        """Property: Set the label of the column."""
        self._label = value

class FieldColumn(Column):
    """Represents a field column in a table.

    Args:
        label (str): The display label of the column.
        field (Property): The property representing the field.
    Attributes:
        label (str): The display label of the column.
        field (Property): The property representing the field.
    """

    def __init__(self, label: str, field: Property):
        super().__init__(label)
        self.field = field

    @property
    def field(self) -> Property:
        """Property: Get the field property of the column."""
        return self._field

    @field.setter
    def field(self, value: Property):
        """Property: Set the field property of the column."""
        self._field = value

    def __repr__(self):
        return f"FieldColumn(label={self.label}, field={self.field.name})"

class LookupColumn(Column):
    """Represents a lookup column in a table.

    Args:
        label (str): The display label of the column.
        path (Property): The property representing the lookup path.
        field (Property): The property representing the lookup field.
        
    Attributes:
        label (str): The display label of the column.
        path (Property): The property representing the lookup path.
        field (Property): The property representing the lookup field.
    """

    def __init__(
        self,
        label: str,
        path: Property,
        field: Property
    ):
        super().__init__(label)
        self.path = path
        self.field = field

    @property
    def path(self) -> Property:
        """Property: Get the path property of the lookup column."""
        return self._path

    @path.setter
    def path(self, value: Property):
        """Property: Set the path property of the lookup column."""
        self._path = value

    @property
    def field(self) -> Property:
        """Property: Get the field property of the lookup column."""
        return self._field

    @field.setter
    def field(self, value: Property):
        """Property: Set the field property of the lookup column."""
        self._field = value

    def __repr__(self):
        return f"LookupColumn(label={self.label}, field={self.field.name})"

class ExpressionColumn(Column):
    """Represents an expression column in a table.

    Args:
        label (str): The display label of the column.
        expression (str): The expression used to compute the column value.
        
    Attributes:
        label (str): The display label of the column.
        expression (str): The expression used to compute the column value.
    """

    def __init__(self, label: str, expression: str):
        super().__init__(label)
        self.expression = expression

    @property
    def expression(self) -> str:
        """Property: Get the expression of the column."""
        return self._expression

    @expression.setter
    def expression(self, value: str):
        """Property: Set the expression of the column."""
        self._expression = value

    def __repr__(self):
        return f"ExpressionColumn(label={self.label}, expression={self.expression})"

class Table(ViewComponent):
    """Represents a table component in the dashboard.

    Args:
        name (str): The name of the table.
        show_header (bool): Whether to render the table header.
        striped_rows (bool): Whether to alternate row background colors.
        show_pagination (bool): Whether to display pagination information.
        rows_per_page (int): Number of rows shown per page.
        title (str | None): Optional title of the table.
        primary_color (str | None): Optional primary color used for the header/background.

    Attributes:
        show_header (bool): Whether to render the table header.
        striped_rows (bool): Whether to alternate row backgrounds.
        show_pagination (bool): Whether to display pagination controls.
        rows_per_page (int): Number of rows per page when pagination is enabled.
    """

    def __init__(
        self,
        name: str,
        show_header: bool = True,
        striped_rows: bool = False,
        show_pagination: bool = False,
        rows_per_page: int = 5,
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        columns: Optional[Sequence[Column]] = None,
        action_buttons: bool = False,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self.title = title
        self.primary_color = primary_color
        self.show_header = show_header
        self.striped_rows = striped_rows
        self.show_pagination = show_pagination
        self.rows_per_page = rows_per_page
        self.columns = list(columns or [])
        self.action_buttons = action_buttons

    @property
    def title(self) -> Optional[str]:
        """Property: Get the title of the table."""
        return self._title

    @title.setter
    def title(self, value: Optional[str]):
        """Property: Set the title of the table."""
        self._title = value

    @property
    def primary_color(self) -> Optional[str]:
        """Property: Get the primary color of the table."""
        return self._primary_color

    @primary_color.setter
    def primary_color(self, value: Optional[str]):
        """Property: Set the primary color of the table."""
        self._primary_color = value

    @property
    def show_header(self) -> bool:
        """Property: Get whether the table header is displayed."""
        return self._show_header

    @show_header.setter
    def show_header(self, value: bool):
        """Property: Set whether the table header is displayed."""
        self._show_header = bool(value)

    @property
    def striped_rows(self) -> bool:
        """Property: Get whether striped rows are enabled."""
        return self._striped_rows

    @striped_rows.setter
    def striped_rows(self, value: bool):
        """Property: Set whether striped rows are enabled."""
        self._striped_rows = bool(value)

    @property
    def show_pagination(self) -> bool:
        """Property: Get whether pagination info is displayed."""
        return self._show_pagination

    @show_pagination.setter
    def show_pagination(self, value: bool):
        """Property: Set whether pagination info is displayed."""
        self._show_pagination = bool(value)

    @property
    def rows_per_page(self) -> int:
        """Property: Get the number of rows per page."""
        return self._rows_per_page

    @rows_per_page.setter
    def rows_per_page(self, value: int):
        """Property: Set the number of rows per page (minimum 1)."""
        try:
            int_value = int(value)
        except (TypeError, ValueError):
            int_value = 5
        self._rows_per_page = max(1, int_value)

    @property
    def columns(self) -> list[Column]:
        """Property: Get the configured column names."""
        return self._columns

    @columns.setter
    def columns(self, value: Sequence[Column]):
        """Property: Set the configured column names."""
        if value is None:
            self._columns = []
            return
        self._columns = [item for item in value if item]

    @property
    def action_buttons(self) -> bool:
        """Property: Get whether action buttons are displayed."""
        return self._action_buttons

    @action_buttons.setter
    def action_buttons(self, value: bool):
        """Property: Set whether action buttons are displayed."""
        self._action_buttons = bool(value)

    def __repr__(self):
        return (
            f"Table(name={self.name}, show_header={self.show_header}, "
            f"striped_rows={self.striped_rows}, rows_per_page={self.rows_per_page}, "
            f"columns={self.columns}, action_buttons={self.action_buttons})"
        )

class MetricCard(ViewComponent):
    """Represents a metric card / KPI card component in the dashboard.
    
    Args:
        name (str): The name of the metric card.
        metric_title (str): The title displayed on the metric card.
        format (str): The format for displaying the value ('number', 'currency', 'percentage', 'time').
        value_color (str): The color of the value text.
        value_size (int): The font size of the value text.
        show_trend (bool): Whether to show the trend indicator.
        positive_color (str): Color for positive trends.
        negative_color (str): Color for negative trends.
        title (str | None): Optional title for the metric card container.
        primary_color (str | None): Optional primary color.
    
    Attributes:
        metric_title (str): The title displayed on the metric card.
        format (str): The format for displaying the value.
        value_color (str): The color of the value text.
        value_size (int): The font size of the value text.
        show_trend (bool): Whether to show the trend indicator.
        positive_color (str): Color for positive trends.
        negative_color (str): Color for negative trends.
    """
    
    def __init__(
        self,
        name: str,
        metric_title: str = "Metric Title",
        format: str = "number",
        value_color: str = "#2c3e50",
        value_size: int = 32,
        show_trend: bool = True,
        positive_color: str = "#27ae60",
        negative_color: str = "#e74c3c",
        title: Optional[str] = None,
        primary_color: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self.metric_title = metric_title
        self.format = format
        self.value_color = value_color
        self.value_size = value_size
        self.show_trend = show_trend
        self.positive_color = positive_color
        self.negative_color = negative_color
        self.title = title
        self.primary_color = primary_color
    
    @property
    def metric_title(self) -> str:
        """Property: Get the metric title."""
        return self._metric_title
    
    @metric_title.setter
    def metric_title(self, value: str):
        """Property: Set the metric title."""
        self._metric_title = value
    
    @property
    def format(self) -> str:
        """Property: Get the display format."""
        return self._format
    
    @format.setter
    def format(self, value: str):
        """Property: Set the display format."""
        self._format = value
    
    @property
    def value_color(self) -> str:
        """Property: Get the value color."""
        return self._value_color
    
    @value_color.setter
    def value_color(self, value: str):
        """Property: Set the value color."""
        self._value_color = value
    
    @property
    def value_size(self) -> int:
        """Property: Get the value font size."""
        return self._value_size
    
    @value_size.setter
    def value_size(self, value: int):
        """Property: Set the value font size."""
        self._value_size = value
    
    @property
    def show_trend(self) -> bool:
        """Property: Get whether to show trend."""
        return self._show_trend
    
    @show_trend.setter
    def show_trend(self, value: bool):
        """Property: Set whether to show trend."""
        self._show_trend = value
    
    @property
    def positive_color(self) -> str:
        """Property: Get the positive trend color."""
        return self._positive_color
    
    @positive_color.setter
    def positive_color(self, value: str):
        """Property: Set the positive trend color."""
        self._positive_color = value
    
    @property
    def negative_color(self) -> str:
        """Property: Get the negative trend color."""
        return self._negative_color
    
    @negative_color.setter
    def negative_color(self, value: str):
        """Property: Set the negative trend color."""
        self._negative_color = value
    
    @property
    def title(self) -> Optional[str]:
        """Optional[str]: Optional title for the metric card container."""
        return self._title
    
    @title.setter
    def title(self, value: Optional[str]):
        """Property: Set the title."""
        self._title = value
    
    @property
    def primary_color(self) -> Optional[str]:
        """Optional[str]: Optional primary color."""
        return self._primary_color
    
    @primary_color.setter
    def primary_color(self, value: Optional[str]):
        """Property: Set the primary color."""
        self._primary_color = value
    
    def __repr__(self):
        return (
            f"MetricCard(name={self.name}, metric_title={self.metric_title}, "
            f"format={self.format}, show_trend={self.show_trend})"
        )

class Map(ViewComponent):
    """Represents a map component in the dashboard.

    Args:
        name (str): The name of the map.

    Attributes:
        name (str): The name of the map.
    """

    def __init__(self, name: str, data: list):
        super().__init__(name)
        self.data = data

class WorldMap(Map):
    """Represents a world map component in the dashboard.

    Args:
        name (str): The name of the world map.
        data (list): The data to be displayed on the world map.

    Attributes:
        name (str): The name of the world map.
        data (list): The data to be displayed on the world map.
    """

    def __init__(self, name: str, data: list):
        super().__init__(name, data)


class LocationMap(Map):
    """Represents a location map component in the dashboard.

    Args:
        name (str): The name of the location map.
        data (list): The data to be displayed on the location map.

    Attributes:
        name (str): The name of the location map.
        data (list): The data to be displayed on the location map.
    """

    def __init__(self, name: str, data: list):
        super().__init__(name, data)



"""
React generator entry point.
Splits rendering and serialization logic into mixins for clarity.
"""
from __future__ import annotations

import os
import shutil
from typing import Any, Dict, List, Tuple

from jinja2 import Environment, FileSystemLoader

from besser.BUML.metamodel.gui import GUIModel
from besser.BUML.metamodel.structural import DomainModel
from besser.generators import GeneratorInterface

from .page_builder import PageBuilderMixin
from .serialization import GuiSerializationMixin


class ReactGenerator(GuiSerializationMixin, PageBuilderMixin, GeneratorInterface):
    """
    Generates React code based on BUML and GUI models.
    Walks templates, renders them with serialized GUI metadata, and builds pages.
    """

    def __init__(self, model: DomainModel, gui_model: GUIModel, output_dir: str = None):
        super().__init__(model, output_dir)
        self.gui_model = gui_model
        templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
        self.env = Environment(
            loader=FileSystemLoader(templates_path),
            trim_blocks=True,
            lstrip_blocks=True,
            extensions=["jinja2.ext.do"],
            variable_start_string="[[",
            variable_end_string="]]",
        )
        self._style_map: Dict[Tuple[str, ...], Dict[str, Any]] = {}
        self._raw_style_entries: List[Dict[str, Any]] = list(getattr(self.gui_model, "_style_entries", []))

    def generate(self):
        """
        Generates React TS code based on the provided BUML and GUI models.
        Generates all files from the templates directory, preserving structure and file names (removing .j2 extension).
        Generates all chart component files (ChartBlock imports them unconditionally).
        """

        context = self._build_generation_context()
        used_component_types = self._get_used_component_types()

        def generate_file_from_template(rel_template_path: str):
            if rel_template_path.endswith(".j2"):
                rel_output_path = rel_template_path[:-3]
            else:
                rel_output_path = rel_template_path

            file_path = self.build_generation_path(file_name=rel_output_path)
            try:
                os.makedirs(os.path.dirname(file_path), exist_ok=True)
                template = self.env.get_template(rel_template_path.replace("\\", "/"))
                generated_code = template.render(**context)
                with open(file_path, mode="w", encoding="utf-8") as f:
                    f.write(generated_code)
            except Exception as exc:
                print(f"Error generating {file_path} from {rel_template_path}: {exc}")
                raise

        templates_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
        for root, _, files in os.walk(templates_path):
            for file_name in files:
                abs_template_path = os.path.join(root, file_name)
                rel_template_path = os.path.relpath(abs_template_path, templates_path)

                if not self._should_generate_file(rel_template_path, used_component_types):
                    continue

                if file_name.endswith(".j2"):
                    generate_file_from_template(rel_template_path)
                else:
                    rel_output_path = rel_template_path
                    dest_path = self.build_generation_path(file_name=rel_output_path)
                    os.makedirs(os.path.dirname(dest_path), exist_ok=True)
                    shutil.copy2(abs_template_path, dest_path)

        self._generate_pages()

    def _should_generate_file(self, rel_path: str, used_component_types: set) -> bool:
        """Determine if a file should be generated based on component usage."""
        if "charts" + os.sep not in rel_path and "table" + os.sep not in rel_path:
            return True

        if "charts" + os.sep in rel_path:
            return True

        component_template_map = {
            "LineChart": ["LineChartComponent.tsx.j2"],
            "BarChart": ["BarChartComponent.tsx.j2"],
            "PieChart": ["PieChartComponent.tsx.j2"],
            "RadarChart": ["RadarChartComponent.tsx.j2"],
            "RadialBarChart": ["RadialBarChartComponent.tsx.j2"],
            "Table": [
                "TableComponent.tsx.j2",
                "TableComponent.css.j2",
                "ColumnFilter.tsx.j2",
                "ColumnSort.tsx.j2",
            ],
            "MetricCard": ["MetricCardComponent.tsx.j2"],
        }

        file_name = os.path.basename(rel_path)
        for component_type, template_files in component_template_map.items():
            if file_name in template_files:
                return component_type in used_component_types

        return True



"""
Project Builder

This module generates Python code for complete BUML projects including all models.
"""

import os
import tempfile
import uuid
import shutil
from besser.BUML.metamodel.structural.structural import DomainModel
from besser.BUML.metamodel.object.object import ObjectModel
from besser.BUML.metamodel.project import Project
from besser.BUML.metamodel.state_machine.agent import Agent
from besser.utilities.buml_code_builder.domain_model_builder import (
    domain_model_to_code,
    contains_user_class,
    is_user_object_model,
)
from besser.utilities.buml_code_builder.agent_model_builder import agent_model_to_code
from besser.utilities.buml_code_builder.quantum_model_builder import quantum_model_to_code

try:
    from besser.utilities.web_modeling_editor.backend.constants.user_buml_model import (
        domain_model as reference_user_domain_model,
    )
except ImportError:
    reference_user_domain_model = None


def project_to_code(project: Project, file_path: str, sm: str = ""):
    """
    Generates Python code for a complete B-UML project and writes it to a specified file.
    
    Parameters:
        project (Project): The B-UML project containing multiple models
        file_path (str): The path where the generated code will be saved
        sm (str): Optional state machine code to include
        
    Outputs:
        - A Python file containing all models in the project
    """
    output_dir = os.path.dirname(file_path)
    if output_dir and not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not file_path.endswith('.py'):
        file_path += '.py'

    domain_model = None
    objectmodel = None
    agent_model = None
    user_domain_model = None
    user_objectmodel = None
    gui_model = None

    # Import GUIModel locally to avoid circular imports
    try:
        from besser.BUML.metamodel.gui import GUIModel
    except ImportError:
        GUIModel = None

    for model in project.models:
        if isinstance(model, DomainModel):
            if contains_user_class(model):
                user_domain_model = user_domain_model or model
            elif domain_model is None:
                domain_model = model
        if isinstance(model, ObjectModel):
            if is_user_object_model(model):
                user_objectmodel = user_objectmodel or model
            elif objectmodel is None:
                objectmodel = model
        if isinstance(model, Agent):
            agent_model = model
        if GUIModel and isinstance(model, GUIModel):
            gui_model = model
            
    # Check for QuantumCircuit model (import locally to avoid circular deps if needed, though less likely here)
    try:
        from besser.BUML.metamodel.quantum import QuantumCircuit
        quantum_model = next((m for m in project.models if isinstance(m, QuantumCircuit)), None)
    except ImportError:
        quantum_model = None

    if user_objectmodel and not user_domain_model and reference_user_domain_model:
        user_domain_model = reference_user_domain_model

    models = []
    with open(file_path, 'w', encoding='utf-8') as f:
        # Models
        temp_dir = tempfile.mkdtemp(prefix=f"besser_{uuid.uuid4().hex}_")

        if objectmodel and domain_model:
            output_file_path = os.path.join(temp_dir, "domain_model.py")
            domain_model_to_code(model=domain_model, file_path=output_file_path, objectmodel=objectmodel)
            with open(output_file_path, "r", encoding='utf-8') as m:
                file_content = m.read()
            content_str = file_content
            f.write(content_str)
            f.write("\n\n")
            models.append("domain_model")
            models.append("object_model")
        elif domain_model:
            output_file_path = os.path.join(temp_dir, "domain_model.py")
            domain_model_to_code(model=domain_model, file_path=output_file_path)
            with open(output_file_path, "r", encoding='utf-8') as m:
                file_content = m.read()
            content_str = file_content
            f.write(content_str)
            f.write("\n\n")
            models.append("domain_model")

        if user_domain_model:
            output_file_path = os.path.join(temp_dir, "user_model.py")
            domain_model_to_code(
                model=user_domain_model,
                file_path=output_file_path,
                objectmodel=user_objectmodel,
                model_var_name="user_model",
                metadata_var_name="user_metadata",
                object_model_var_name="user_object_model",
            )
            with open(output_file_path, "r", encoding='utf-8') as m:
                file_content = m.read()
            content_str = file_content
            f.write(content_str)
            f.write("\n\n")
            models.append("user_model")
            if user_objectmodel:
                models.append("user_object_model")
        
        if agent_model:
            output_file_path = os.path.join(temp_dir, "agent_model.py")
            agent_model_to_code(model=agent_model, file_path=output_file_path)
            with open(output_file_path, "r", encoding='utf-8') as m:
                file_content = m.read()
            content_str = file_content
            f.write(content_str)
            f.write("\n\n")
            models.append("agent")
        
        if gui_model:
            # Use gui_code_builder to generate GUI model code
            # gui_model_to_code appends to file, but we're already writing, so write to temp file first
            output_file_path = os.path.join(temp_dir, "gui_model.py")
            from besser.utilities.buml_code_builder.gui_model_builder import gui_model_to_code
            # Pass domain_model=None because we already wrote it above
            gui_model_to_code(model=gui_model, file_path=output_file_path, domain_model=None)
            with open(output_file_path, "r", encoding='utf-8') as m:
                file_content = m.read()
            content_str = file_content
            f.write(content_str)
            f.write("\n\n")
            models.append("gui_model")

        if quantum_model:
            output_file_path = os.path.join(temp_dir, "quantum_model.py")
            quantum_model_to_code(model=quantum_model, file_path=output_file_path)
            with open(output_file_path, "r", encoding='utf-8') as m:
                file_content = m.read()
            content_str = file_content
            f.write(content_str)
            f.write("\n\n")
            models.append("quantum_model")

        if sm != "":
            f.write(sm)
            models.append("sm")
            f.write("\n\n")

        shutil.rmtree(temp_dir, ignore_errors=True)

        # Project
        # Write imports
        f.write("######################\n")
        f.write("# PROJECT DEFINITION #\n")
        f.write("######################\n\n")
        f.write("from besser.BUML.metamodel.project import Project\n")
        f.write("from besser.BUML.metamodel.structural.structural import Metadata\n\n")

        # Write metadata and project
        f.write(f'metadata = Metadata(description="{project.metadata.description}")\n')
        f.write("project = Project(\n")
        f.write(f'    name="{project.name}",\n')
        f.write(f'    models=[{", ".join(models)}],\n')
        f.write(f'    owner="{project.owner}",\n')
        f.write("    metadata=metadata\n")
        f.write(")\n")
    
    print(f"Project saved to {file_path}")



from .utils import *
from .image_to_buml import *
from .buml_code_builder import *
from .buml_code_builder.domain_model_builder import domain_model_to_code
from .buml_code_builder.agent_model_builder import agent_model_to_code
from .buml_code_builder.gui_model_builder import gui_model_to_code
from .buml_code_builder.project_builder import project_to_code


def __getattr__(name: str):
    """
    Lazy-load kg_to_buml symbols to avoid import-time circular dependencies.
    """
    try:
        from . import kg_to_buml as _kg_to_buml
    except Exception as exc:  # pragma: no cover - defensive fallback
        raise AttributeError(name) from exc
    if hasattr(_kg_to_buml, name):
        return getattr(_kg_to_buml, name)
    raise AttributeError(name)



import pytest
from besser.BUML.metamodel.object import *
from besser.BUML.metamodel.structural import (
    Class, Property, Method, Parameter,
    BinaryAssociation, Generalization, DomainModel,
    Enumeration, EnumerationLiteral, Multiplicity,
    StringType, IntegerType, FloatType, BooleanType,
    TimeType, DateType, DateTimeType, TimeDeltaType,
    AnyType, Constraint, AssociationClass
)

# Domain model definition

# Classes
Person = Class(name="Person")
Player = Class(name="Player")
Team = Class(name="Team")
Coach = Class(name="Coach")

# Person class attributes and methods
Person_country: Property = Property(name="country", type=StringType)
Person.attributes={Person_country}

# Player class attributes and methods
Player_age: Property = Property(name="age", type=IntegerType)
Player_jerseyNumber: Property = Property(name="jerseyNumber", type=IntegerType)
Player_position: Property = Property(name="position", type=StringType)
Player_name: Property = Property(name="name", type=StringType)
Player.attributes={Player_jerseyNumber, Player_position, Player_name, Player_age}

# Team class attributes and methods
Team_name: Property = Property(name="name", type=StringType)
Team_city: Property = Property(name="city", type=StringType)
Team_division: Property = Property(name="division", type=StringType)
Team.attributes={Team_city, Team_name, Team_division}

# Coach class attributes and methods
Coach_name: Property = Property(name="name", type=StringType)
Coach_salary: Property = Property(name="salary", type=IntegerType)
Coach.attributes={Coach_name, Coach_salary}

# Relationships
person_team_asso: BinaryAssociation = BinaryAssociation(
    name="person_team_asso",
    ends={
        Property(name="follows", type=Team, multiplicity=Multiplicity(1, 9999)),
        Property(name="fans", type=Person, multiplicity=Multiplicity(0, 9999))
    }
)
team_player_asso: BinaryAssociation = BinaryAssociation(
    name="team_player_asso",
    ends={
        Property(name="plays_for", type=Team, multiplicity=Multiplicity(1, 1)),
        Property(name="players", type=Player, multiplicity=Multiplicity(0, 9999))
    }
)
Team_Coach: BinaryAssociation = BinaryAssociation(
    name="Team_Coach",
    ends={
        Property(name="leads", type=Team, multiplicity=Multiplicity(1, 1)),
        Property(name="lead_by", type=Coach, multiplicity=Multiplicity(1, 9999))
    }
)

# Generalizations
gen_Player_Person = Generalization(general=Person, specific=Player)
gen_Coach_Person = Generalization(general=Person, specific=Coach)

# Domain Model
domain_model = DomainModel(
    name="Team_Playermodel",
    types={Person, Player, Team, Coach},
    associations={person_team_asso, team_player_asso, Team_Coach},
    generalizations={gen_Player_Person, gen_Coach_Person}
)

# Tests for the fluent API of the object model

def test_team_creation():
    team = Team("team1").attributes(name="Lakers", city="LA", division="West").build()
    assert team.name == "Lakers"
    assert team.city == "LA"
    assert team.division == "West"

def test_player_creation():
    player = Player("player1").attributes(name="LeBron", age=38, position="Forward", jerseyNumber=6, country="USA").build()
    assert player.name == "LeBron"
    assert player.age == 38
    assert player.position == "Forward"
    assert player.jerseyNumber == 6
    assert player.country == "USA"

def test_coach_creation():
    coach = Coach("coach1").attributes(name="Vogel", salary=100, country="USA").build()
    assert coach.name == "Vogel"
    assert coach.salary == 100
    assert coach.country == "USA"

def test_player_plays_for_team():
    team = Team("team2").attributes(name="Warriors", city="SF", division="West").build()
    player = Player("player2").attributes(name="Curry", age=34, position="Guard", jerseyNumber=30, country="USA") \
        .link(team, "plays_for").build()
    assert player.plays_for == team
    assert player.plays_for.division == "West"

def test_player_follows_team():
    team = Team("team3").attributes(name="Bulls", city="Chicago", division="East").build()
    player = Player("player3").attributes(name="Jordan", age=59, position="Guard", jerseyNumber=23, country="USA") \
        .link(team, "follows").build()
    assert player.follows == team
    assert player.follows.city == "Chicago"

def test_coach_leads_team():
    team = Team("team4").attributes(name="Celtics", city="Boston", division="East").build()
    coach = Coach("coach2").attributes(name="Stevens", salary=120, country="USA") \
        .link(team, "leads").build()
    assert coach.leads == team
    assert coach.leads.city == "Boston"

def test_coach_follows_team():
    team = Team("team5").attributes(name="Heat", city="Miami", division="East").build()
    team2 = Team("team2").attributes(name="Test", city="Lux", division="East").build()
    coach = Coach("coach3").attributes(name="Spoelstra", salary=110, country="USA") \
        .link({team,team2}, "follows").build()
    assert team in coach.follows
    assert any(t.city == "Miami" for t in coach.follows)

def test_object_model_creation():
    team = Team("team6").attributes(name="Knicks", city="NY", division="East").build()
    player = Player("player4").attributes(name="Barrett", age=22, position="Guard", jerseyNumber=9, country="Canada").build()
    coach = Coach("coach4").attributes(name="Thibodeau", salary=90, country="USA").build()
    obj_model = ObjectModel(name="NBA_Model", objects={team, player, coach})
    assert team in obj_model.objects
    assert player in obj_model.objects
    assert coach in obj_model.objects

def test_multiple_links():
    team1 = Team("team8").attributes(name="Nets", city="Brooklyn", division="East").build()
    team2 = Team("team9").attributes(name="Raptors", city="Toronto", division="East").build()
    player = Player("player6").attributes(name="Durant", age=33, position="Forward", jerseyNumber=7, country="USA") \
        .link(team1, "plays_for") \
        .link({team1, team2}, "follows").build()
    assert player.plays_for == team1
    assert team2 in player.follows
    assert team1 in player.follows

def test_update_attributes_after_build():
    team = Team("team10").attributes(name="Spurs", city="San Antonio", division="West").build()
    team.city = "Austin"
    assert team.city == "Austin"

def test_object_internal_name():
    team = Team("team11").attributes(name="Kings", city="Sacramento", division="West").build()
    assert team.name == "Kings"
    assert team.name_ == "team11"
    team.name_ = "new_team_internal_name"
    assert team.name_ == "new_team_internal_name"
    assert team.name == "Kings"


from abc import ABC
from typing import TypeVar
from typing import TYPE_CHECKING

from besser.BUML.metamodel.structural import Method, Property, Class, Enumeration

if TYPE_CHECKING:
    from besser.BUML.metamodel.action_language.visitors import BALVisitor


ContextType = TypeVar('ContextType')
ReturnType = TypeVar('ReturnType')

############################################
# Definition of Classes
############################################


#==================#
# Abstract Classes #
#==================#


class Statement(ABC):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Statements(self, context)
    

class Expression(Statement):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Expression(self, context)


class AssignTarget(ABC):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_AssignTarget(self, context)


#===============#
# Types Classes #
#===============#


class Type(ABC):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Type(self, context)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __le__(self, other):
        if isinstance(other, Type):
            return self.is_subtype(other)
        else:
            return False

    def __ge__(self, other):
        if isinstance(other, Type):
            return other.is_subtype(self)
        else:
            return False

    def __or__(self, other):
        if isinstance(other, Type):
            return TypeUnion(self, other)
        else:
            return self

    def supertype(self):
        return self

    def subtype(self):
        return self

    # check if a is a subtype of b (a<=b)
    def is_subtype(self, other):
        if not isinstance(other, Type):
            return False

        a = self.supertype()
        b = other.supertype()
        if a == b:
            return True
        if a == Nothing():
            return True
        if b == AnyType():
            return True
        if isinstance(a, TypeUnion):
            if isinstance(a.a, Nothing):
                return a.b <= b
            else:
                return a.a <= b
        if isinstance(b, TypeUnion):
            if isinstance(b.a, Nothing):
                return a <= b.b
            else:
                return a <= b.a
        elif isinstance(a, OptionalType):
            if isinstance(b, OptionalType):
                return a.type <= b.type
            else:
                return a.type <= b
        elif isinstance(b, OptionalType):
            return a <= b.type
        elif isinstance(a, FunctionType) and isinstance(b, FunctionType):
            # P1 -> R1 <= P2 -> R2 ==>  P2 <= P1, R1 <= R2
            if not a.return_type <= b.return_type:  # if ~ R1 <= R2
                return False

            shortest, longest = (a.params_type, b.params_type) \
                if len(a.params_type) <= len(b.params_type) \
                else (b.params_type, a.params_type)

            for i in range(len(shortest), len(longest)):
                if not isinstance(longest[i], OptionalType):
                    return False

            for i in range(0, len(shortest)):
                if not b.params_type[i] <= a.params_type[i]:  # if ~ P2 <= P1
                    return False
            return True
        elif (not isinstance(a, ObjectType)) or (not isinstance(b, ObjectType)):
            return False
        else:  # Both are ObjectType
            b_children: set[Class] = b.clazz.all_specializations()
            if a.clazz in b_children:
                return True
            return False


class NoType(Type):
    pass

class Nothing(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Nothing(self, context)

class TypeUnion(Type):
    def __init__(self, a: Type, b: Type):
        self.a = a
        self.b = b

    def supertype(self) -> Type:
        #Find the supertype R of the union (A | B) <= R
        # (A | B) <= R <==> A <= R, B <= R
        a = self.a.supertype()
        b = self.b.supertype()
        if a == b:
            return a
        elif isinstance(a, NoType):
            return b
        elif isinstance(b, NoType):
            return a
        elif isinstance(a, Nothing):
            if isinstance(b, Nothing):
                return Nothing()
            else:
                return self
        elif isinstance(b, Nothing):
            return b | a
        elif isinstance(a, TypeUnion):
            if isinstance(a.a, Nothing):
                union_type = a.b | b
                return a.a | union_type.supertype()
            else:
                union_type = a.a | b
                return a.b | union_type.supertype()
        elif isinstance(b, TypeUnion):
            if isinstance(b.a, Nothing):
                union_type = b.b | a
                return b.a | union_type.supertype()
            else:
                union_type = b.a | a
                return b.b | union_type.supertype()
        elif isinstance(a, OptionalType):
            if isinstance(b, OptionalType):
                union = a.type | b.type
            else:
                union = a.type | b
            return OptionalType(union.supertype())
        elif isinstance(b, OptionalType):
            union = a | b.type
            return OptionalType(union.supertype())
        elif isinstance(a, FunctionType) and isinstance(b, FunctionType):
            # (P1 -> R1 | P2 -> R2) <= P -> R ==> P <= P1, P <= P2, R1 <= R, R2 <= R
            # equivalent to (P1 -> R1 | P2 -> R2) <= P -> R ==> P <= (P1 | P2), (R1 | R2) <= R
            return_type = (a.return_type | b.return_type).supertype() # compute R such as (R1 | R2) <= R
            shortest, longest = (a.params_type, b.params_type) \
                                if len(a.params_type) <= len(b.params_type)\
                                else (b.params_type, a.params_type)
            for i in range(len(shortest), len(longest)):
                if not isinstance(longest[i], OptionalType):
                    return AnyType()
            param_types = []
            for i in range(0, len(shortest)):
                union = a.params_type[i] | b.params_type[i]
                # function params are contravariant -> the union of param types are common subtypes
                param_type = union.subtype()
                if param_type is None:
                    # No possible type found -> fallback to any
                    return AnyType()
                param_types.append(param_type)
            return FunctionType(param_types, return_type)
        elif (not isinstance(a, ObjectType)) or (not isinstance(b, ObjectType)):
            return AnyType()
        else: # Both are ObjectType
            a_parents: set[Class] = a.clazz.all_parents()
            b_parents: set[Class] = b.clazz.all_parents()
            if a.clazz in b_parents:
                return a
            if b.clazz in a_parents:
                return b

            common_parent: set[Class] = a_parents.intersection(b_parents)

            while True:
                to_remove: set[Class] = set()
                for parent in common_parent:
                    to_remove.union(parent.all_parents().intersection(common_parent))
                common_parent.difference_update(to_remove)
                if len(to_remove) == 0:
                    break
            for clazz in common_parent:
                return ObjectType(clazz)
            return AnyType()

    def subtype(self) -> Type:
        # Find the subtype R of the union R <= (A | B)
        # R <= (A | B) <==> R <= A, R <= B
        a = self.a.subtype()
        b = self.b.subtype()
        if isinstance(a, NoType):
            return NoType()
        elif isinstance(b, NoType):
            return NoType()
        elif a == AnyType():
            return b
        elif b == AnyType():
            return a
        elif a == b:
            return a
        elif isinstance(a, Nothing):
            if isinstance(b, Nothing):
                return Nothing()
            else:
                return self
        elif isinstance(b, Nothing):
            return b | a
        elif isinstance(a, TypeUnion):
            if isinstance(a.a, Nothing):
                union_type = a.b | b
                return a.a | union_type.subtype()
            else:
                union_type = a.a | b
                return a.b | union_type.subtype()
        elif isinstance(b, TypeUnion):
            if isinstance(b.a, Nothing):
                union_type = b.b | a
                return b.a | union_type.subtype()
            else:
                union_type = b.a | a
                return b.b | union_type.subtype()
        elif isinstance(a, OptionalType):
            if isinstance(b, OptionalType):
                union = (a.type | b.type).subtype()
                if union is not None:
                    return OptionalType(union)
                return NoType()
            else:
                return (a.type | b).subtype()
        elif isinstance(b, OptionalType):
            union = a | b.type
            return union.subtype()
        elif isinstance(a, FunctionType) and isinstance(b, FunctionType):
            # P -> R <= (P1 -> R1 | P2 -> R2) ==> P1 <= P, P2 <= P, R <= R1, R <= R2
            # equivalent to P -> R <= (P1 -> R1 | P2 -> R2) ==> (P1 | P2) <= P, R <= (R1 | R2)
            return_type = (a.return_type | b.return_type).subtype() # compute R such as  R <= (R1 | R2)
            if return_type == NoType(): # there is no R such as R <= (R1 | R2)
                return NoType()

            shortest, longest = (a.params_type, b.params_type) \
                                if len(a.params_type) <= len(b.params_type) \
                                else (b.params_type, a.params_type)

            for i in range(len(shortest), len(longest)):
                if not isinstance(longest[i], OptionalType):
                    return NoType()

            param_types = []
            for i in range(0, len(shortest)):
                param_type = (a.params_type[i] | b.params_type[i]).supertype() # compute P such as (P1 | P2) <= P
                param_types.append(param_type)
            return FunctionType(param_types, return_type)
        elif (not isinstance(a, ObjectType)) or (not isinstance(b, ObjectType)):
            return NoType()
        else:  # Both are ObjectType
            a_children: set[Class] = a.clazz.all_specializations()
            b_children: set[Class] = b.clazz.all_specializations()
            if a.clazz in b_children:
                return a
            if b.clazz in a_children:
                return b

            common_children: set[Class] = a_children.intersection(b_children)

            while True:
                to_remove: set[Class] = set()
                for child in common_children:
                    to_remove.union(child.all_specializations().intersection(common_children))
                common_children.difference_update(to_remove)
                if len(to_remove) == 0:
                    break
            for clazz in common_children:
                return ObjectType(clazz)
            return NoType()


class AnyType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_AnyType(self, context)


class ObjectType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_ObjectType(self, context)

    def __init__(self, clazz: 'Class' = None):
        self.__clazz = clazz

    def __eq__(self, other):
        if isinstance(other, ObjectType):
            return self.__clazz.name == other.__clazz.name
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def clazz(self) -> 'Class':
        return self.__clazz

    @clazz.setter
    def clazz(self, clazz: 'Class'):
        self.__clazz = clazz


class SequenceType(Type):
    def __init__(self, elementsType: Type):
        self.__elementsType = elementsType

    def __eq__(self, other):
        if isinstance(other, SequenceType):
            return self.__elementsType == other.__elementsType
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def elementsType(self) -> Type:
        return self.__elementsType

    @elementsType.setter
    def elementsType(self, elementsType: Type):
        self.__elementsType = elementsType

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_SequenceType(self, context)


class BoolType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_BoolType(self, context)


class StringType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_StringType(self, context)


class RealType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_RealType(self, context)


class IntType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_IntType(self, context)


class EnumType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_EnumType(self, context)

    def __init__(self, enum: 'Enumeration' = None):
        self.__enum = enum

    def __eq__(self, other):
        if isinstance(other, EnumType):
            return self.__enum == other.__enum
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def enum(self) -> 'Enumeration':
        return self.__enum

    @enum.setter
    def enum(self, enum: 'Enumeration'):
        self.__enum = enum


# Type of parameters with default
class OptionalType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_OptionalType(self, context)

    def __init__(self, type: Type):
        self.__type = type

    @property
    def type(self) -> Type:
        return self.__type

    @type.setter
    def type(self, type: Type):
        self.__type = type

    def __eq__(self, other):
        if isinstance(other, OptionalType):
            return self.__type == other.__type
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class FunctionType(Type):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_FunctionType(self, context)

    def __init__(self, params_type: list[Type], return_type: Type):
        self.__params_type = params_type
        self.__return_type = return_type

    def __eq__(self, other):
        if isinstance(other, FunctionType):
            if self.__return_type != other.__return_type:
                return False
            if len(self.__params_type) != len(other.__params_type):
                return False
            for i in range(0, len(self.__params_type)):
                if self.__params_type[i] != other.__params_type[i]:
                    return False
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def params_type(self) -> list[Type]:
        return self.__params_type

    @params_type.setter
    def params_type(self, params_type: list[Type]):
        self.__params_type = params_type

    @property
    def return_type(self) -> Type:
        return self.__return_type

    @return_type.setter
    def return_type(self, return_type: Type):
        self.__return_type = return_type


#==========================#
# Name Declaration Classes #
#==========================#


class Multiplicity:
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Multiplicity(self, context)

    def __init__(self, nullable: bool, many: bool):
        self.__nullable = nullable
        self.__many = many

    @property
    def many(self) -> bool:
        return self.__many

    @many.setter
    def many(self, many: bool):
        self.__many = many

    @property
    def nullable(self) -> bool:
        return self.__nullable

    @nullable.setter
    def nullable(self, nullable: bool):
        self.__nullable = nullable


class NameDecl(Statement):
    def __init__(self, name: str, declared_type: Type, multiplicity: Multiplicity):
        self.__name = name
        self.__declared_type = declared_type
        self.__multiplicity = multiplicity

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name

    @property
    def declared_type(self) -> Type:
        return self.__declared_type

    @declared_type.setter
    def declared_type(self, declared_type: Type):
        self.__declared_type = declared_type

    @property
    def multiplicity(self) -> Multiplicity:
        return self.__multiplicity

    @multiplicity.setter
    def multiplicity(self, multiplicity: Multiplicity):
        self.__multiplicity = multiplicity

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_NameDecl(self, context)


class ExplicitDecl(AssignTarget, NameDecl):
    def __init__(self, name: str, declared_type: Type, multiplicity: Multiplicity):
        super().__init__(name, declared_type, multiplicity)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_ExplicitDecl(self, context)


class ImplicitDecl(AssignTarget, NameDecl):
    def __init__(self, name: str, declared_type: Type, multiplicity: Multiplicity):
        super().__init__(name, declared_type, multiplicity)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_ImplicitDecl(self, context)
    

class Parameter(NameDecl):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Parameter(self, context)

    def __init__(self, name: str, declared_type: "Type" = None, default: "Expression" = None):
        mult = Multiplicity(True, isinstance(declared_type, SequenceType))
        super().__init__(name, declared_type, mult)
        self.__default = default

    @property
    def default(self):
        return self.__default

    @default.setter
    def default(self, value):
        self.__default = value


#=====================#
# Function Definition #
#      Root Class     #
#=====================#


class FunctionDefinition(NameDecl):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_FunctionDefinition(self, context)

    def __init__(self, name: str, parameters: list["Parameter"] = None, return_type: "Type" = None,
                 body: "Block" = None):
        self.__parameters = parameters if parameters is not None else set()
        self.__return_type = return_type
        self.__body = body
        param_types = [param.declared_type for param in self.__parameters]
        super().__init__(name, FunctionType(param_types, return_type), Multiplicity(False, False))



    @property
    def return_type(self):
        return self.__return_type

    @return_type.setter
    def return_type(self, value):
        self.__return_type = value

    @property
    def parameters(self):
        return self.__parameters

    @parameters.setter
    def parameters(self, value):
        self.__parameters = value if value is not None else set()

    @property
    def body(self):
        return self.__body

    @body.setter
    def body(self, value):
        self.__body = value


#====================#
# Statements Classes #
#====================#


class ConditionalBranch(ABC):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_ConditionalBranch(self, context)
    

class Block(ConditionalBranch):
    def __init__(self, statements:list[Statement]):
        self.__statements = statements

    @property
    def statements(self) -> list[Statement]:
        return self.__statements

    @statements.setter
    def statements(self, statements: list[Statement]):
        self.__statements = statements

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Block(self, context)


class Condition(Statement, ConditionalBranch):
    def __init__(self, condition: Expression, then: Block, elze: ConditionalBranch):
        self.__condition = condition
        self.__then = then
        self.__elze = elze

    @property
    def condition(self) -> Expression:
        return self.__condition

    @condition.setter
    def condition(self, condition: Expression):
        self.__condition = condition

    @property
    def then(self) -> Block:
        return self.__then

    @then.setter
    def then(self, then: Block):
        self.__then = then

    @property
    def elze(self) -> ConditionalBranch:
        return self.__elze

    @elze.setter
    def elze(self, elze: ConditionalBranch):
        self.__elze = elze

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Condition(self, context)
    

class CondLoop(Statement):
    def __init__(self, condition: Expression, body: Block):
        self.__condition = condition
        self.__body = body

    @property
    def condition(self) -> any:
        return self.__condition

    @condition.setter
    def condition(self, condition: Expression):
        self.__condition = condition

    @property
    def body(self) -> any:
        return self.__body

    @body.setter
    def body(self, body: Block):
        self.__body = body

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_CondLoop(self, context)


class DoWhile(CondLoop):
    def __init__(self, condition: Expression, body: Block):
        super().__init__(condition, body)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_DoWhile(self, context)


class While(CondLoop):
    def __init__(self, condition: Expression, body: Block):
        super().__init__(condition, body)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_While(self, context)


class Iterator:
    def __init__(self, var_name: NameDecl, sequence: Expression):
        self.__var_name = var_name
        self.__sequence = sequence

    @property
    def var_name(self) -> NameDecl:
        return self.__var_name

    @var_name.setter
    def var_name(self, var_name: NameDecl):
        self.__var_name = var_name

    @property
    def sequence(self) -> Expression:
        return self.__sequence

    @sequence.setter
    def sequence(self, sequence: Expression):
        self.__sequence = sequence

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Iterator(self, context)


class For(Statement):
    def __init__(self, iterators:set[Iterator], body:Block):
        self.__iterators = iterators
        self.__body = body

    @property
    def iterators(self) -> set[Iterator]:
        return self.__iterators

    @iterators.setter
    def iterators(self, iterators: set[Iterator]):
        self.__iterators = iterators

    @property
    def body(self) -> Block:
        return self.__body

    @body.setter
    def body(self, body: Block):
        self.__body = body

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_For(self, context)


class Assignment(Statement):
    def __init__(self, target: AssignTarget, assignee: Expression):
        self.__target = target
        self.__assignee = assignee

    @property
    def target(self) -> any:
        return self.__target

    @target.setter
    def target(self, target: any):
        self.__target = target

    @property
    def assignee(self) -> any:
        return self.__assignee

    @assignee.setter
    def assignee(self, assignee: any):
        self.__assignee = assignee

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Assignment(self, context)


class Return(Statement):
    def __init__(self, expr: Expression):
        self.__expr = expr

    @property
    def expr(self) -> Expression:
        return self.__expr

    @expr.setter
    def expr(self, expr: Expression):
        self.__expr = expr

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Return(self, context)


#=====================#
# Expressions Classes #
#=====================#


class Ternary(Expression):
    def __init__(self, expr: Expression, then: Expression, elze: Expression):
        self.__expr = expr
        self.__then = then
        self.__elze = elze

    @property
    def expr(self) -> Expression:
        return self.__expr

    @expr.setter
    def expr(self, expr: Expression):
        self.__expr = expr

    @property
    def then(self):
        return self.__then

    @then.setter
    def then(self, value):
        self.__then = value

    @property
    def elze(self) -> Expression:
        return self.__elze

    @elze.setter
    def elze(self, elze: Expression):
        self.__elze = elze

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Ternary(self, context)


# Boolean Expressions #
class Boolean(Expression):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Boolean(self, context)


class Not(Boolean):
    def __init__(self, expr: Expression):
        self.__expr = expr

    @property
    def expr(self):
        return self.__expr

    @expr.setter
    def expr(self, value):
        self.__expr = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Not(self, context)


class BinaryBoolean(Boolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        self.__left = left
        self.__right = right

    @property
    def left(self) -> Expression:
        return self.__left

    @left.setter
    def left(self, left: Expression):
        self.__left = left

    @property
    def right(self):
        return self.__right

    @right.setter
    def right(self, value):
        self.__right = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_BinaryBoolean(self, context)


class Or(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Or(self, context)


class And(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_And(self, context)


class Unequal(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Unequal(self, context)


class Equal(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Equal(self, context)


class GreaterEq(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_GreaterEq(self, context)


class LessEq(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_LessEq(self, context)


class Greater(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Greater(self, context)


class Less(BinaryBoolean):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Less(self, context)


class InstanceOf(Expression):
    def __init__(self, instance: Expression, the_type: Type):
        self.__instance = instance
        self.__type = the_type

    @property
    def instance(self) -> Expression:
        return self.__instance

    @instance.setter
    def instance(self, instance: Expression):
        self.__instance = instance

    @property
    def type(self):
        return self.__type

    @type.setter
    def type(self, value):
        self.__type = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_InstanceOf(self, context)


# Arithmetic Expressions #
class Concatenation(Expression):
    def __init__(self, left: Expression, right: "Expression" = None):
        self.__left = left
        self.__right = right

    @property
    def left(self) -> Expression:
        return self.__left

    @left.setter
    def left(self, left: Expression):
        self.__left = left

    @property
    def right(self):
        return self.__right

    @right.setter
    def right(self, value):
        self.__right = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Concatenation(self, context)


class Arithmetic(Expression):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Arithmetic(self, context)


class UnaryMinus(Arithmetic):
    def __init__(self, expr: Expression):
        self.__expr = expr

    @property
    def expr(self):
        return self.__expr

    @expr.setter
    def expr(self, value):
        self.__expr = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_UnaryMinus(self, context)


class BinaryArithmetic(Arithmetic):
    def __init__(self, left: Expression, right: "Expression" = None):
        self.__left = left
        self.__right = right

    @property
    def left(self) -> Expression:
        return self.__left

    @left.setter
    def left(self, left: Expression):
        self.__left = left

    @property
    def right(self):
        return self.__right

    @right.setter
    def right(self, value):
        self.__right = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_BinaryArithmetic(self, context)


class Plus(BinaryArithmetic):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Plus(self, context)


class Minus(BinaryArithmetic):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Minus(self, context)


class Mult(BinaryArithmetic):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Mult(self, context)


class Div(BinaryArithmetic):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Div(self, context)


class Remain(BinaryArithmetic):
    def __init__(self, left: Expression, right: "Expression" = None):
        super().__init__(left, right)

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Remain(self, context)


# Type Expressions #
class Cast(Expression):
    def __init__(self, instance: Expression, as_type: Type):
        self.__instance = instance
        self.__as_type = as_type

    @property
    def instance(self) -> Expression:
        return self.__instance

    @instance.setter
    def instance(self, instance: Expression):
        self.__instance = instance

    @property
    def as_type(self):
        return self.__as_type

    @as_type.setter
    def as_type(self, value):
        self.__as_type = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Cast(self, context)


class NullCoalessing(Expression):
    def __init__(self, nullable: Expression, elze: "Expression" = None):
        self.__nullable = nullable
        self.__elze = elze

    @property
    def nullable(self) -> Expression:
        return self.__nullable

    @nullable.setter
    def nullable(self, nullable: Expression):
        self.__nullable = nullable

    @property
    def elze(self):
        return self.__elze

    @elze.setter
    def elze(self, value):
        self.__elze = value

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_NullCoalessing(self, context)


# Access Expressions #
class FieldAccess(AssignTarget, Expression):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_FieldAccess(self, context)

    def __init__(self, receiver: Expression, field: 'Property'):
        self.__field = field
        self.__receiver = receiver

    @property
    def field(self) -> 'Property':
        return self.__field

    @field.setter
    def field(self, field: 'Property'):
        self.__field = field

    @property
    def receiver(self):
        return self.__receiver

    @receiver.setter
    def receiver(self, value):
        self.__receiver = value

class ArrayAccess(AssignTarget, Expression):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_ArrayAccess(self, context)

    def __init__(self, receiver: "Expression", index: Expression):
        self.__index = index
        self.__receiver = receiver

    @property
    def index(self) -> Expression:
        return self.__index

    @index.setter
    def index(self, index: Expression):
        self.__index = index

    @property
    def receiver(self):
        return self.__receiver

    @receiver.setter
    def receiver(self, value):
        self.__receiver = value


# Call Expressions #
class Call(Expression):
    def __init__(self, arguments: list[Expression]):
        self.__arguments = arguments

    @property
    def arguments(self) -> list[Expression]:
        return self.__arguments

    @arguments.setter
    def arguments(self, arguments: list[Expression]):
        self.__arguments = arguments

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Call(self, context)


class MethodCall(Call):
    def __init__(self, receiver: Expression, method: 'Method', arguments: list[Expression]):
        super().__init__(arguments)
        self.__receiver = receiver
        self.__method = method

    @property
    def receiver(self) -> Expression:
        return self.__receiver

    @receiver.setter
    def receiver(self, receiver: Expression):
        self.__receiver = receiver

    @property
    def method(self) -> 'Method':
        return self.__method

    @method.setter
    def method(self, method: 'Method'):
        self.__method = method

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_MethodCall(self, context)


class New(Call):
    def __init__(self, clazz: ObjectType, arguments: list[Expression]):
        super().__init__(arguments)
        self.__clazz = clazz

    @property
    def clazz(self) -> ObjectType:
        return self.__clazz

    @clazz.setter
    def clazz(self, clazz: ObjectType):
        self.__clazz = clazz

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_New(self, context)

class StandardLibCall(Call):
    def __init__(self, receiver: Expression, function: str, fn_type: FunctionType, arguments: list[Expression]):
        super().__init__(arguments)
        self.__receiver = receiver
        self.__function = function
        self.__function_type = fn_type

    @property
    def receiver(self) -> Expression:
        return self.__receiver

    @receiver.setter
    def receiver(self, receiver: Expression):
        self.__receiver = receiver

    @property
    def function(self) -> str:
        return self.__function

    @function.setter
    def function(self, function: str):
        self.__function = function

    @property
    def function_type(self) -> FunctionType:
        return self.__function_type

    @function_type.setter
    def function_type(self, function_type: FunctionType):
        self.__function_type = function_type

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_StandardLibCall(self, context)


class ProcedureCall(Call):
    def __init__(self, function: FunctionDefinition, arguments: list[Expression]):
        super().__init__(arguments)
        self.__function = function

    @property
    def function(self) -> FunctionDefinition:
        return self.__function

    @function.setter
    def function(self, function: FunctionDefinition):
        self.__function = function

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_ProcedureCall(self, context)


# References Expressions #
class This(Expression):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_This(self, context)

    def __init__(self, clazz: Class):
        self.__clazz = clazz

    @property
    def clazz(self) -> Class:
        return self.__clazz

    @clazz.setter
    def clazz(self, clazz: Class):
        self.__clazz = clazz


class Reference(AssignTarget, Expression):
    def __init__(self, definition: NameDecl, symbol:str = None):
        self.__definition = definition

    @property
    def definition(self) -> NameDecl:
        return self.__definition

    @definition.setter
    def definition(self, definition: NameDecl):
        self.__definition = definition

    @property
    def symbol(self) -> str:
        return self.__symbol

    @symbol.setter
    def symbol(self, symbol: str):
        self.__symbol = symbol

    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Reference(self, context)


#==================#
# Literals Classes #
#==================#


class Literal(Expression):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_Literal(self, context)


class StringLiteral(Literal):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_StringLiteral(self, context)

    def __init__(self, value: str):
        self.__value = value

    @property
    def value(self) -> str:
        return self.__value

    @value.setter
    def value(self, value: str):
        self.__value = value


class BoolLiteral(Literal):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_BoolLiteral(self, context)

    def __init__(self, value: bool):
        self.__value = value

    @property
    def value(self) -> bool:
        return self.__value

    @value.setter
    def value(self, value: bool):
        self.__value = value


class RealLiteral(Literal):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_RealLiteral(self, context)

    def __init__(self, value: float):
        self.__value = value

    @property
    def value(self) -> float:
        return self.__value

    @value.setter
    def value(self, value: float):
        self.__value = value


class IntLiteral(Literal):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_IntLiteral(self, context)

    def __init__(self, value: int):
        self.__value = value

    @property
    def value(self) -> int:
        return self.__value

    @value.setter
    def value(self, value: int):
        self.__value = value


class NullLiteral(Literal):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_NullLiteral(self, context)


class EnumLiteral(Literal):
    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_EnumLiteral(self, context)

    def __init__(self, enumeration: EnumType, name: str):
        self.__name = name
        self.__enumeration = enumeration

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name

    @property
    def enumeration(self):
        return self.__enumeration

    @enumeration.setter
    def enumeration(self, value):
        self.__enumeration = value


class SequenceLiteral(Literal):
    def __init__(self, value_type: Type, values: list[Expression]):
        self.__value_type = value_type
        self.__values = values


    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_SequenceLiteral(self, context)

    @property
    def value_type(self) -> Type:
        return self.__value_type

    @value_type.setter
    def value_type(self, value_type: Type):
        self.__value_type = value_type

    @property
    def values(self) -> list[Expression]:
        return self.__values

    @values.setter
    def values(self, values: list[Expression]):
        self.__values = values


class RangeLiteral(Literal):
    def __init__(self, first: Expression, last: Expression):
        self.__first = first
        self.__last = last

    @property
    def first(self) -> Expression:
        return self.__first

    @first.setter
    def first(self, first: Expression):
        self.__first = first

    @property
    def last(self) -> Expression:
        return self.__last

    @last.setter
    def last(self, last: Expression):
        self.__last = last


    def accept(self, bal_visitor: 'BALVisitor[ContextType, ReturnType]', context: ContextType) -> ReturnType:
        return bal_visitor.visit_RangeLiteral(self, context)


from besser.BUML.metamodel.structural import NamedElement, Model, Metadata

class Project(NamedElement):
    """A class representing a project containing multiple models.
    Args:
        name (str): The name of the project.
        models (list[Model], optional): A list of models contained in the project. Defaults to an empty list.
        metadata (Metadata, optional): Metadata associated with the project. Defaults to None.

    Attributes:
        name (str): The name of the project.
        models (list[Model]): A list of models contained in the project.
        metadata (Metadata): Metadata associated with the project.
    """
    def __init__(self, name: str, models: list[Model] = None, owner: str = "", metadata: Metadata = None):
        super().__init__(name)
        self.models = models if models is not None else []
        self.owner = owner
        self.metadata = metadata

    @property
    def models(self) -> list[Model]:
        """list[Model]: Get the list of models."""
        return self.__models

    @models.setter
    def models(self, models: list[Model]):
        """list[Model]: Set the list of models."""
        if not isinstance(models, list):
            raise TypeError("Models must be a list of Model instances.")
        self.__models = models

    @property
    def owner(self) -> str:
        """str: Get the owner of the project."""
        return self.__owner

    @owner.setter
    def owner(self, owner: str):
        """str: Set the owner of the project."""
        self.__owner = owner



from __future__ import annotations

import json
import os
import re
from typing import Any, Dict, List, Optional, Set


class _PageRenderContext:
    def __init__(self) -> None:
        self.imports: Set[str] = set()
        self.needs_navigate: bool = False


class PageBuilderMixin:
    def _generate_pages(self):
        """Build real TSX pages and App.tsx, and remove legacy renderer artifacts."""
        components_payload, _, meta = self._serialize_gui_model()
        pages = components_payload.get("pages", [])
        self._page_path_map = self._build_page_path_map(pages)
        selector_style_map = self._build_selector_style_map()

        page_infos = self._write_page_components(pages, selector_style_map)
        self._write_app_tsx(page_infos, meta)

    def _build_selector_style_map(self) -> Dict[str, Dict[str, Any]]:
        selector_style_map: Dict[str, Dict[str, Any]] = {}
        for entry in self._style_map.values():
            for selector in entry.get("selectors", []) or []:
                if selector not in selector_style_map:
                    selector_style_map[selector] = entry.get("style", {}) or {}
        return selector_style_map

    def _write_page_components(
        self, pages: List[Dict[str, Any]], selector_style_map: Dict[str, Dict[str, Any]]
    ) -> List[Dict[str, Any]]:
        pages_dir = self.build_generation_path("src/pages")
        os.makedirs(pages_dir, exist_ok=True)

        used_names: Set[str] = set()
        page_infos: List[Dict[str, Any]] = []

        for idx, page in enumerate(pages):
            raw_name = page.get("name") or page.get("id") or f"Page{idx + 1}"
            component_name = self._to_pascal_case(raw_name)
            if not component_name:
                component_name = f"Page{idx + 1}"
            if component_name[0].isdigit():
                component_name = f"Page{component_name}"

            base_name = component_name
            counter = 1
            while component_name in used_names:
                counter += 1
                component_name = f"{base_name}{counter}"
            used_names.add(component_name)

            context = _PageRenderContext()
            jsx = self._render_page_jsx(page, selector_style_map, context)

            imports = self._build_page_imports(context)
            navigate_hook = "  const navigate = useNavigate();\n" if context.needs_navigate else ""

            file_contents = (
                "import React from \"react\";\n"
                + ("import { useNavigate } from \"react-router-dom\";\n" if context.needs_navigate else "")
                + imports
                + "\n"
                + f"const {component_name}: React.FC = () => {{\n"
                + navigate_hook
                + "  return (\n"
                + jsx
                + "  );\n"
                + "};\n\n"
                + f"export default {component_name};\n"
            )

            file_path = os.path.join(pages_dir, f"{component_name}.tsx")
            with open(file_path, "w", encoding="utf-8") as f:
                f.write(file_contents)

            page_infos.append(
                {
                    "component_name": component_name,
                    "file_name": f"./pages/{component_name}",
                    "route_path": page.get("route_path") or "/",
                    "is_main": bool(page.get("is_main")),
                    "page_id": page.get("id"),
                }
            )

        return page_infos

    @staticmethod
    def _build_page_path_map(pages: List[Dict[str, Any]]) -> Dict[str, str]:
        page_path_map: Dict[str, str] = {}
        for page in pages or []:
            page_id = page.get("id")
            route_path = page.get("route_path")
            if not route_path:
                name = page.get("name") or page_id
                route_path = PageBuilderMixin._path_from_name(name)
            if not route_path:
                continue
            if page_id:
                page_path_map[page_id] = route_path
                page_path_map[f"page:{page_id}"] = route_path
            if page.get("name"):
                page_path_map[str(page["name"])] = route_path
        return page_path_map

    def _build_page_imports(self, context: _PageRenderContext) -> str:
        imports: List[str] = []
        if "ChartBlock" in context.imports:
            imports.append("import { ChartBlock } from \"../components/runtime/ChartBlock\";")
        if "TableBlock" in context.imports:
            imports.append("import { TableBlock } from \"../components/runtime/TableBlock\";")
        if "MetricCardBlock" in context.imports:
            imports.append("import { MetricCardBlock } from \"../components/runtime/MetricCardBlock\";")
        if "DataListBlock" in context.imports:
            imports.append("import { DataListBlock } from \"../components/runtime/DataListBlock\";")
        if "AgentComponent" in context.imports:
            imports.append("import { AgentComponent } from \"../components/AgentComponent\";")
        if "MethodButton" in context.imports:
            imports.append("import { MethodButton } from \"../components/MethodButton\";")

        if not imports:
            return ""
        return "\n".join(imports) + "\n"

    def _render_page_jsx(
        self,
        page: Dict[str, Any],
        selector_style_map: Dict[str, Dict[str, Any]],
        context: _PageRenderContext,
    ) -> str:
        page_id = page.get("id") or ""
        components = page.get("components") or []

        page_style = self._compute_style(
            component_id=page_id,
            class_list=[],
            selector_style_map=selector_style_map,
            has_children=bool(components),
        )

        props = self._build_element_props(
            component_id=page_id,
            class_list=[],
            style=page_style,
            attributes={},
            on_click_expr=None,
        )

        children_jsx = self._render_component_list(components, selector_style_map, context, indent=4)
        if children_jsx:
            return (
                f"    <div{props}>\n"
                + children_jsx
                + "    </div>\n"
            )
        return f"    <div{props} />\n"

    def _render_component_list(
        self,
        components: List[Dict[str, Any]],
        selector_style_map: Dict[str, Dict[str, Any]],
        context: _PageRenderContext,
        indent: int,
    ) -> str:
        rendered = []
        for component in components:
            rendered.append(
                self._render_component(component, selector_style_map, context, indent)
            )
        return "\n".join([line for line in rendered if line])

    def _render_component(
        self,
        node: Dict[str, Any],
        selector_style_map: Dict[str, Dict[str, Any]],
        context: _PageRenderContext,
        indent: int,
    ) -> str:
        comp_type = node.get("type") or "component"
        component_id = node.get("id") or ""
        class_list = node.get("class_list") or []
        attributes = dict(node.get("attributes") or {})
        children = node.get("children") or []

        style = self._compute_style(
            component_id=component_id,
            class_list=class_list,
            selector_style_map=selector_style_map,
            has_children=bool(children),
        )

        indent_str = " " * indent

        # Containers
        if comp_type in {"container", "wrapper", "component"}:
            tag = self._select_container_tag(node.get("tag"))
            props = self._build_element_props(component_id, class_list, style, attributes, None)
            if children:
                children_jsx = self._render_component_list(children, selector_style_map, context, indent + 2)
                return (
                    f"{indent_str}<{tag}{props}>\n"
                    + children_jsx
                    + f"\n{indent_str}</{tag}>"
                )
            return f"{indent_str}<{tag}{props} />"

        # Text
        if comp_type == "text":
            tag = self._select_text_tag(node.get("tag"))
            content = node.get("content") or ""
            content_expr = f"{{{json.dumps(content, ensure_ascii=False)}}}"
            props = self._build_element_props(component_id, class_list, style, attributes, None)
            return f"{indent_str}<{tag}{props}>{content_expr}</{tag}>"

        # Button
        if comp_type in {"button", "action-button"}:
            return self._render_button(node, component_id, class_list, style, attributes, context, indent_str)

        # Link
        if comp_type in {"link", "link-button"}:
            url = node.get("url") or ""
            target_path = self._resolve_target_path(node)
            on_click_expr = None
            if target_path and (not url or url == "#" or not str(url).startswith("http")):
                context.needs_navigate = True
                on_click_expr = (
                    f"(e) => {{ e.preventDefault(); navigate({json.dumps(target_path)}); }}"
                )
                url = target_path
            if not url or url == "#":
                url = target_path or "/"

            link_attrs = dict(attributes)
            link_attrs.pop("href", None)
            props = self._build_element_props(
                component_id,
                class_list,
                style,
                link_attrs,
                on_click_expr,
                extra_props={
                    "href": url,
                    "target": node.get("target") or None,
                    "rel": node.get("rel") or None,
                },
            )
            label = node.get("label") or node.get("name") or ""
            content_expr = f"{{{json.dumps(label, ensure_ascii=False)}}}"
            return f"{indent_str}<a{props}>{content_expr}</a>"

        # Image
        if comp_type == "image":
            img_attrs = dict(attributes)
            img_attrs.pop("src", None)
            img_attrs.pop("alt", None)
            props = self._build_element_props(
                component_id,
                class_list,
                style,
                img_attrs,
                None,
                extra_props={
                    "src": node.get("src") or "",
                    "alt": node.get("alt") or node.get("description") or "",
                },
            )
            return f"{indent_str}<img{props} />"

        # Input
        if comp_type == "input":
            input_type = (node.get("input_type") or attributes.get("type") or "text")
            input_attrs = dict(attributes)
            input_attrs.pop("type", None)
            props = self._build_element_props(
                component_id,
                class_list,
                style,
                input_attrs,
                None,
                extra_props={
                    "type": str(input_type).lower(),
                    "placeholder": attributes.get("placeholder") or "",
                },
            )
            return f"{indent_str}<input{props} />"

        # Form
        if comp_type == "form":
            on_submit_expr = "(e) => { e.preventDefault(); }"
            props = self._build_element_props(component_id, class_list, style, attributes, None)
            props += " onSubmit={" + on_submit_expr + "}"
            inputs = node.get("inputs") or []
            input_lines = []
            for input_field in inputs:
                input_id = input_field.get("id") or ""
                label = input_field.get("label") or input_id
                field_type = (input_field.get("type") or "text").lower()
                label_expr = f"{{{json.dumps(label, ensure_ascii=False)}}}"
                input_lines.append(
                    f"{indent_str}  <div style={{{{\"marginBottom\": \"10px\"}}}}>\n"
                    f"{indent_str}    <label htmlFor={json.dumps(input_id)} style={{{{\"display\": \"block\", \"marginBottom\": \"5px\"}}}}>"
                    f"{label_expr}"
                    f"</label>\n"
                    f"{indent_str}    <input id={json.dumps(input_id)} type={json.dumps(field_type)} />\n"
                    f"{indent_str}  </div>"
                )
            inner = "\n".join(input_lines) if input_lines else ""
            return (
                f"{indent_str}<form{props}>\n"
                + inner
                + f"\n{indent_str}  <button type=\"submit\">Submit</button>\n"
                + f"{indent_str}</form>"
            )

        # Menu
        if comp_type == "menu":
            items = node.get("items") or []
            props = self._build_element_props(component_id, class_list, style, attributes, None)
            menu_list_style = self._format_prop(
                "style",
                {"listStyle": "none", "padding": 0, "margin": 0},
            )
            list_item_style = self._format_prop(
                "style",
                {"display": "inline-block", "marginRight": "15px"},
            )
            item_lines = []
            for item in items:
                label = item.get("label") or ""
                url = item.get("url") or ""
                if not url or url == "#":
                    url = self._page_path_map.get(label) or "/"
                target = item.get("target")
                rel = item.get("rel")
                label_expr = f"{{{json.dumps(label, ensure_ascii=False)}}}"
                item_lines.append(
                    f"{indent_str}    <li {list_item_style}>"
                    f"<a href={json.dumps(url)}"
                    + (f" target={json.dumps(target)}" if target else "")
                    + (f" rel={json.dumps(rel)}" if rel else "")
                    + f">{label_expr}</a></li>"
                )
            items_jsx = "\n".join(item_lines)
            return (
                f"{indent_str}<nav{props}>\n"
                f"{indent_str}  <ul {menu_list_style}>\n"
                + items_jsx
                + f"\n{indent_str}  </ul>\n"
                + f"{indent_str}</nav>"
            )

        # Data List
        if comp_type == "data-list":
            context.imports.add("DataListBlock")
            props = self._build_component_props(
                component_id=component_id,
                class_list=class_list,
                style=style,
                extra_props={
                    "dataSources": node.get("data_sources") or [],
                },
            )
            return f"{indent_str}<DataListBlock{props} />"

        # Embedded content (iframe)
        if comp_type == "embedded-content":
            iframe_attrs = dict(attributes)
            iframe_attrs.pop("src", None)
            props = self._build_element_props(
                component_id,
                class_list,
                style,
                iframe_attrs,
                None,
                extra_props={
                    "src": node.get("src") or "",
                    "title": node.get("description") or node.get("name") or "Embedded content",
                },
            )
            return f"{indent_str}<iframe{props} />"

        # Charts
        if comp_type in {"bar-chart", "line-chart", "pie-chart", "radar-chart", "radial-bar-chart"}:
            context.imports.add("ChartBlock")
            props = self._build_component_props(
                component_id=component_id,
                class_list=class_list,
                style=style,
                extra_props={
                    "chartType": comp_type,
                    "title": node.get("title"),
                    "color": node.get("color"),
                    "chart": node.get("chart"),
                    "series": node.get("series"),
                    "dataBinding": node.get("data_binding"),
                },
                style_prop_name="styles",
                include_class_name=False,
            )
            return f"{indent_str}<ChartBlock{props} />"

        # Table
        if comp_type == "table":
            context.imports.add("TableBlock")
            props = self._build_component_props(
                component_id=component_id,
                class_list=class_list,
                style=style,
                extra_props={
                    "title": node.get("title"),
                    "options": node.get("chart"),
                    "dataBinding": node.get("data_binding"),
                },
                style_prop_name="styles",
                include_class_name=False,
            )
            return f"{indent_str}<TableBlock{props} />"

        # Metric card
        if comp_type == "metric-card":
            context.imports.add("MetricCardBlock")
            props = self._build_component_props(
                component_id=component_id,
                class_list=class_list,
                style=style,
                extra_props={
                    "metric": node.get("metric"),
                    "dataBinding": node.get("data_binding"),
                },
                style_prop_name="styles",
                include_class_name=False,
            )
            return f"{indent_str}<MetricCardBlock{props} />"

        # Agent component
        if comp_type == "agent-component":
            context.imports.add("AgentComponent")
            props = self._build_component_props(
                component_id=component_id,
                class_list=class_list,
                style=style,
                extra_props={
                    "agent-name": node.get("agent-name") or node.get("agent_name") or "",
                    "agent-title": node.get("agent-title") or node.get("agent_title") or "BESSER Agent",
                },
                style_prop_name="styles",
                include_class_name=False,
            )
            return f"{indent_str}<AgentComponent{props} />"

        # Fallback
        tag = self._select_container_tag(node.get("tag"))
        props = self._build_element_props(component_id, class_list, style, attributes, None)
        if children:
            children_jsx = self._render_component_list(children, selector_style_map, context, indent + 2)
            return (
                f"{indent_str}<{tag}{props}>\n"
                + children_jsx
                + f"\n{indent_str}</{tag}>"
            )
        return f"{indent_str}<{tag}{props} />"

    def _render_button(
        self,
        node: Dict[str, Any],
        component_id: str,
        class_list: List[str],
        style: Dict[str, Any],
        attributes: Dict[str, Any],
        context: _PageRenderContext,
        indent_str: str,
    ) -> str:
        action_type = node.get("action_type")
        if action_type == "run-method" or "endpoint" in attributes:
            context.imports.add("MethodButton")
            endpoint = attributes.get("endpoint")
            is_instance_method = attributes.get("is-instance-method")
            instance_source = attributes.get("instance-source")
            input_params = attributes.get("input-parameters") or {}
            parameters: List[Dict[str, Any]] = []
            if isinstance(input_params, dict):
                for name, param_meta in input_params.items():
                    if isinstance(param_meta, dict):
                        parameter: Dict[str, Any] = {
                            "name": name,
                            "type": str(param_meta.get("type", "any")),
                            "required": bool(param_meta.get("required", True)),
                        }
                        if "default" in param_meta:
                            parameter["default"] = param_meta.get("default")
                        if param_meta.get("input_kind"):
                            parameter["inputKind"] = str(param_meta.get("input_kind"))
                        if param_meta.get("entity"):
                            parameter["entity"] = str(param_meta.get("entity"))
                        if param_meta.get("lookup_field"):
                            parameter["lookupField"] = str(param_meta.get("lookup_field"))
                        if isinstance(param_meta.get("options"), list):
                            parameter["options"] = param_meta.get("options")
                        parameters.append(parameter)
                    else:
                        parameters.append({"name": name, "type": str(param_meta), "required": True})
            elif isinstance(input_params, list):
                for item in input_params:
                    if not isinstance(item, dict):
                        continue
                    name = item.get("name")
                    if not name:
                        continue
                    parameter: Dict[str, Any] = {
                        "name": name,
                        "type": str(item.get("type", "any")),
                        "required": bool(item.get("required", True)),
                    }
                    if "default" in item:
                        parameter["default"] = item.get("default")
                    if item.get("inputKind"):
                        parameter["inputKind"] = str(item.get("inputKind"))
                    if item.get("entity"):
                        parameter["entity"] = str(item.get("entity"))
                    if item.get("lookupField"):
                        parameter["lookupField"] = str(item.get("lookupField"))
                    if isinstance(item.get("options"), list):
                        parameter["options"] = item.get("options")
                    parameters.append(parameter)
            label = attributes.get("button-label") or node.get("label") or node.get("name") or "Execute"

            props = self._build_component_props(
                component_id=component_id,
                class_list=class_list,
                style=style,
                extra_props={
                    "endpoint": endpoint,
                    "label": label,
                    "parameters": parameters if parameters else None,
                    "isInstanceMethod": self._to_bool(is_instance_method),
                    "instanceSourceTableId": instance_source,
                },
            )
            return f"{indent_str}<MethodButton{props} />"

        target_path = self._resolve_target_path(node)
        on_click_expr = None
        if target_path:
            context.needs_navigate = True
            on_click_expr = f"() => navigate({json.dumps(target_path)})"

        props = self._build_element_props(component_id, class_list, style, attributes, on_click_expr)
        label = node.get("label") or node.get("name") or "Button"
        content_expr = f"{{{json.dumps(label, ensure_ascii=False)}}}"
        return f"{indent_str}<button{props}>{content_expr}</button>"

    # ------------------------------------------------------------------ #
    # Prop helpers
    # ------------------------------------------------------------------ #
    def _build_element_props(
        self,
        component_id: str,
        class_list: List[str],
        style: Dict[str, Any],
        attributes: Dict[str, Any],
        on_click_expr: Optional[str],
        extra_props: Optional[Dict[str, Any]] = None,
    ) -> str:
        props: List[str] = []
        if component_id:
            props.append(f" id={json.dumps(component_id)}")

        class_names = [cls[1:] if cls.startswith(".") else cls for cls in (class_list or []) if cls]
        if class_names:
            props.append(f" className={json.dumps(' '.join(class_names))}")

        if style:
            props.append(f" {self._format_prop('style', style)}")

        if on_click_expr:
            props.append(" onClick={" + on_click_expr + "}")

        cleaned_attrs = dict(attributes or {})
        for key in ("id", "class", "className", "style"):
            cleaned_attrs.pop(key, None)

        if extra_props:
            for key, value in extra_props.items():
                if value is not None and value != "":
                    props.append(f" {self._format_prop(key, value)}")

        if cleaned_attrs:
            props.append(f" {{...{json.dumps(cleaned_attrs, ensure_ascii=False)}}}")

        return "".join(props)

    def _build_component_props(
        self,
        component_id: str,
        class_list: List[str],
        style: Dict[str, Any],
        extra_props: Dict[str, Any],
        style_prop_name: str = "style",
        include_class_name: bool = True,
    ) -> str:
        props: List[str] = []
        if component_id:
            props.append(f" id={json.dumps(component_id)}")

        if include_class_name:
            class_names = [cls[1:] if cls.startswith(".") else cls for cls in (class_list or []) if cls]
            if class_names:
                props.append(f" className={json.dumps(' '.join(class_names))}")

        if style:
            props.append(f" {self._format_prop(style_prop_name, style)}")

        for key, value in (extra_props or {}).items():
            if value is None or value == "":
                continue
            props.append(f" {self._format_prop(key, value)}")

        return "".join(props)

    @staticmethod
    def _format_prop(name: str, value: Any) -> str:
        if isinstance(value, bool):
            return f"{name}={{{str(value).lower()}}}"
        if isinstance(value, (int, float)):
            return f"{name}={{{value}}}"
        if isinstance(value, (list, dict)):
            return f"{name}={{{json.dumps(value, ensure_ascii=False)}}}"
        return f"{name}={json.dumps(value, ensure_ascii=False)}"

    # ------------------------------------------------------------------ #
    # Style helpers for page rendering
    # ------------------------------------------------------------------ #
    def _compute_style(
        self,
        component_id: str,
        class_list: List[str],
        selector_style_map: Dict[str, Dict[str, Any]],
        has_children: bool,
    ) -> Dict[str, Any]:
        style: Dict[str, Any] = {}
        if component_id:
            selector = f"#{component_id}"
            if selector in selector_style_map:
                style.update(selector_style_map[selector])

        for cls in class_list or []:
            selector = cls if cls.startswith((".", "#")) else f".{cls}"
            if selector in selector_style_map:
                style.update(selector_style_map[selector])

        if "gjs-cell" in (class_list or []) and has_children:
            if "height" in style:
                style["height"] = "auto"

        if "gjs-row" in (class_list or []):
            if style.get("display") == "table":
                style["display"] = "flex"
                style.setdefault("flexWrap", "wrap")

        if "gjs-cell" in (class_list or []):
            if style.get("display") in ("table-cell", "block"):
                style.pop("display", None)
            width = style.get("width")
            if isinstance(width, str) and width.strip().endswith("%"):
                style.pop("width", None)
            style.setdefault("flex", "1 1 calc(33.333% - 20px)")
            style.setdefault("minWidth", "250px")

        return style

    # ------------------------------------------------------------------ #
    # Navigation helpers
    # ------------------------------------------------------------------ #
    def _resolve_target_path(self, node: Dict[str, Any]) -> Optional[str]:
        if node.get("target_screen_path"):
            return node.get("target_screen_path")
        target_screen_id = node.get("target_screen_id")
        if target_screen_id:
            mapped = getattr(self, "_page_path_map", {}).get(target_screen_id)
            if mapped:
                return mapped

        events = node.get("events") or []
        for event in events:
            actions = event.get("actions") or []
            for action in actions:
                if action.get("kind") == "Transition":
                    target_path = action.get("target_screen_path")
                    if not target_path:
                        target_id = action.get("target_screen_id") or action.get("target_screen")
                        if target_id:
                            target_path = getattr(self, "_page_path_map", {}).get(target_id)
                    return target_path or self._path_from_name(action.get("target_screen"))
        return None

    @staticmethod
    def _path_from_name(name: Optional[str]) -> Optional[str]:
        if not name:
            return None
        return "/" + str(name).strip().lower().replace(" ", "-")

    # ------------------------------------------------------------------ #
    # Tag helpers
    # ------------------------------------------------------------------ #
    @staticmethod
    def _select_container_tag(tag: Optional[str]) -> str:
        valid_tags = {
            "section",
            "article",
            "header",
            "footer",
            "nav",
            "aside",
            "main",
            "div",
            "ul",
            "ol",
            "li",
            "span",
        }
        tag_value = (tag or "").lower()
        return tag_value if tag_value in valid_tags else "div"

    @staticmethod
    def _select_text_tag(tag: Optional[str]) -> str:
        valid_tags = {"h1", "h2", "h3", "h4", "h5", "h6", "p", "span", "div", "strong", "em", "li"}
        tag_value = (tag or "").lower()
        return tag_value if tag_value in valid_tags else "p"

    # ------------------------------------------------------------------ #
    # Utility helpers for pages
    # ------------------------------------------------------------------ #
    @staticmethod
    def _to_pascal_case(value: str) -> str:
        if not value:
            return ""
        parts = re.split(r"[^a-zA-Z0-9]+", value)
        return "".join(part.capitalize() for part in parts if part)

    @staticmethod
    def _to_bool(value: Any) -> Optional[bool]:
        if value is None:
            return None
        if isinstance(value, bool):
            return value
        if isinstance(value, str):
            return value.strip().lower() == "true"
        return bool(value)

    # ------------------------------------------------------------------ #
    # App.tsx generation for page routing
    # ------------------------------------------------------------------ #
    def _write_app_tsx(self, page_infos: List[Dict[str, Any]], meta: Dict[str, Any]) -> None:
        if not page_infos:
            return

        default_route = None
        main_page_id = meta.get("main_page_id")
        for page in page_infos:
            if main_page_id and page.get("page_id") == main_page_id:
                default_route = page.get("route_path")
                break
        if not default_route:
            main_page = next((p for p in page_infos if p.get("is_main")), None)
            default_route = (main_page or page_infos[0]).get("route_path")

        import_lines = []
        for page in page_infos:
            import_lines.append(
                f"import {page['component_name']} from \"{page['file_name']}\";"
            )
        imports = "\n".join(import_lines)

        routes = "\n".join(
            [
                f"            <Route path={json.dumps(p['route_path'])} element={{<{p['component_name']} />}} />"
                for p in page_infos
            ]
        )

        app_contents = (
            "import React from \"react\";\n"
            "import { Routes, Route, Navigate } from \"react-router-dom\";\n"
            "import { TableProvider } from \"./contexts/TableContext\";\n"
            + imports
            + "\n\n"
            "function App() {\n"
            "  return (\n"
            "    <TableProvider>\n"
            "      <div className=\"app-container\">\n"
            "        <main className=\"app-main\">\n"
            "          <Routes>\n"
            + routes
            + "\n"
            f"            <Route path=\"/\" element={{<Navigate to={json.dumps(default_route)} replace />}} />\n"
            f"            <Route path=\"*\" element={{<Navigate to={json.dumps(default_route)} replace />}} />\n"
            "          </Routes>\n"
            "        </main>\n"
            "      </div>\n"
            "    </TableProvider>\n"
            "  );\n"
            "}\n"
            "export default App;\n"
        )

        app_path = self.build_generation_path("src/App.tsx")
        with open(app_path, "w", encoding="utf-8") as f:
            f.write(app_contents)
    # --------------------------------------------------------------------- #
    # Context builders
    # --------------------------------------------------------------------- #




from besser.BUML.metamodel.action_language.action_language import Condition, Block, ConditionalBranch, DoWhile, While, \
    CondLoop, Iterator, RangeLiteral, SequenceLiteral, EnumLiteral, NullLiteral, RealLiteral, BoolLiteral, \
    StringLiteral, IntLiteral, Literal, Reference, InstanceOf, Concatenation, Minus, Plus, Mult, Remain, Div, \
    BinaryArithmetic, UnaryMinus, Arithmetic, Ternary, NullCoalessing, Cast, ArrayAccess, FieldAccess, This, \
    ProcedureCall, StandardLibCall, MethodCall, New, Call, Not, GreaterEq, Unequal, Less, Greater, Equal, Or, And, \
    LessEq, BinaryBoolean, Boolean, Expression, Return, For, ImplicitDecl, ExplicitDecl, NameDecl, Statement, \
    Assignment, BoolType, EnumType, IntType, StringType, RealType, Type, FunctionType, SequenceType, ObjectType, \
    Multiplicity, AnyType, OptionalType, FunctionDefinition, Parameter, AssignTarget, Nothing
from besser.BUML.metamodel.action_language.visitors import BALVisitor

def bal_to_python(method: FunctionDefinition) -> str:
    generator = BALPythonGenerator()
    return generator.generate(method)

class PythonGenerationContext:
    pass

def indent(lines):
    out = []
    for line in lines:
        out.append('    '+line)
    return out

class BALPythonGenerator(BALVisitor[PythonGenerationContext, list[str]]):

    def generate(self, node: FunctionDefinition):
        lines = node.accept(self, PythonGenerationContext())
        return "\n".join(lines)

    def visit_AssignTarget(self, node: AssignTarget, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_Parameter(self, node: Parameter, context: PythonGenerationContext) -> list[str]:
        if node.declared_type is None:
            return [node.name]
        param_type = node.declared_type.accept(self, context)[0]
        expr = ""
        if node.default is not None:
            expr = node.default.accept(self, context)[0]
            expr = " = " + expr
        return [f"{node.name}: {param_type}{expr}"]

    def visit_FunctionDefinition(self, node: FunctionDefinition, context: PythonGenerationContext) -> list[str]:
        lines = []

        params = []
        for param in node.parameters:
            params += param.accept(self, context)
        lines.append(f"def {node.name}({', '.join(params)}) -> {node.return_type.accept(self, context)[0]} :")

        body = node.body.accept(self, context)
        lines.extend(indent(body))

        return lines


    def visit_AnyType(self, node: AnyType, context: PythonGenerationContext) -> list[str]:
        return ["Any"]

    def visit_Multiplicity(self, node: Multiplicity, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_ObjectType(self, node: ObjectType, context: PythonGenerationContext) -> list[str]:
        return [node.clazz.name]

    def visit_SequenceType(self, node: SequenceType, context: PythonGenerationContext) -> list[str]:
        return [f"list[{node.elementsType.accept(self, context)[0]}]"]

    def visit_OptionalType(self, node: OptionalType, context: PythonGenerationContext) -> list[str]:
        return node.type.accept(self, context)

    def visit_FunctionType(self, node: FunctionType, context: PythonGenerationContext) -> list[str]:
        return ["Any"]

    def visit_Type(self, node: Type, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_RealType(self, node: RealType, context: PythonGenerationContext) -> list[str]:
        return ["float"]

    def visit_StringType(self, node: StringType, context: PythonGenerationContext) -> list[str]:
        return ["str"]

    def visit_IntType(self, node: IntType, context: PythonGenerationContext) -> list[str]:
        return ["int"]

    def visit_EnumType(self, node: EnumType, context: PythonGenerationContext) -> list[str]:
        return [node.enum.name]

    def visit_BoolType(self, node: BoolType, context: PythonGenerationContext) -> list[str]:
        return ["bool"]

    def visit_Nothing(self, node: Nothing, context: PythonGenerationContext) -> list[str]:
        return ["None"]

    def visit_Assignment(self, node: Assignment, context: PythonGenerationContext) -> list[str]:
        target = node.target.accept(self, context)
        assignee = node.assignee.accept(self, context)
        return [f"{target[0]} = {assignee[0]}"]

    def visit_Statement(self, node: Statement, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_NameDecl(self, node: NameDecl, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_ExplicitDecl(self, node: ExplicitDecl, context: PythonGenerationContext) -> list[str]:
        return [f"{node.name}: {node.declared_type.accept(self, context)[0]}"]

    def visit_ImplicitDecl(self, node: ImplicitDecl, context: PythonGenerationContext) -> list[str]:
        return [f"{node.name}"]

    def visit_For(self, node: For, context: PythonGenerationContext) -> list[str]:
        lines = []
        body = []
        for iterator in node.iterators:
            expr = iterator.sequence.accept(self, context)[0]
            name = iterator.var_name.name + "_" + str(node.__hash__())
            lines.append(f"{name} = {expr}")
            body.append(f"{iterator.var_name.name} = {next(iter(node.iterators)).var_name.name}_{node.__hash__()}[i]")

        lines.append(f"for i in range(0, len({next(iter(node.iterators)).var_name.name}_{node.__hash__()})):")
        body.extend(node.body.accept(self, context))
        lines.extend(indent(body))

        return lines

    def visit_Return(self, node: Return, context: PythonGenerationContext) -> list[str]:
        expr = node.expr.accept(self, context)[0]
        return [f"return {expr}"]

    def visit_Expression(self, node: Expression, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_Boolean(self, node: Boolean, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_BinaryBoolean(self, node: BinaryBoolean, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_LessEq(self, node: LessEq, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} <= {right})"]

    def visit_And(self, node: And, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} and {right})"]

    def visit_Or(self, node: Or, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} or {right})"]

    def visit_Equal(self, node: Equal, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} == {right})"]

    def visit_Greater(self, node: Greater, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} > {right})"]

    def visit_Less(self, node: Less, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} < {right})"]

    def visit_Unequal(self, node: Unequal, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} != {right})"]

    def visit_GreaterEq(self, node: GreaterEq, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} >= {right})"]

    def visit_Not(self, node: Not, context: PythonGenerationContext) -> list[str]:
        expr = node.expr.accept(self, context)[0]
        return [f"(not {expr})"]

    def visit_Call(self, node: Call, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_New(self, node: New, context: PythonGenerationContext) -> list[str]:
        args = []
        for arg in node.arguments:
            args.append(arg.accept(self, context)[0])
        return [f"{node.clazz.clazz.name}({', '.join(args)})"]

    def visit_MethodCall(self, node: MethodCall, context: PythonGenerationContext) -> list[str]:
        args = []
        for arg in node.arguments:
            args.append(arg.accept(self, context)[0])
        receiver = node.receiver.accept(self, context)[0]
        return [f"{receiver}.{node.method.name}({', '.join(args)})"]

    def visit_StandardLibCall(self, node: StandardLibCall, context: PythonGenerationContext) -> list[str]:
        args = []
        for arg in node.arguments:
            args.append(arg.accept(self, context)[0])
        receiver = node.receiver.accept(self, context)[0]
        return [f"{node.function}({receiver}, {', '.join(args)})"]

    def visit_ProcedureCall(self, node: ProcedureCall, context: PythonGenerationContext) -> list[str]:
        args = []
        for arg in node.arguments:
            args.append(arg.accept(self, context)[0])
        return [f"{node.function.name}({', '.join(args)})"]

    def visit_This(self, node: This, context: PythonGenerationContext) -> list[str]:
        return ["self"]

    def visit_FieldAccess(self, node: FieldAccess, context: PythonGenerationContext) -> list[str]:
        receiver = node.receiver.accept(self, context)[0]
        return [f"{receiver}.{node.field.name}"]

    def visit_ArrayAccess(self, node: ArrayAccess, context: PythonGenerationContext) -> list[str]:
        receiver = node.receiver.accept(self, context)[0]
        index = node.index.accept(self, context)[0]
        return [f"{receiver}[{index}]"]

    def visit_Cast(self, node: Cast, context: PythonGenerationContext) -> list[str]:
        return node.instance.accept(self, context)

    def visit_NullCoalessing(self, node: NullCoalessing, context: PythonGenerationContext) -> list[str]:
        nullable = node.nullable.accept(self, context)
        elze = node.elze.accept(self, context)
        return [f"({nullable} if {nullable} is not None else {elze})"]

    def visit_Ternary(self, node: Ternary, context: PythonGenerationContext) -> list[str]:
        expr = node.expr.accept(self, context)
        then = node.then.accept(self, context)
        elze = node.elze.accept(self, context)
        return [f"({then} if {expr} else {elze})"]

    def visit_Arithmetic(self, node: Arithmetic, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_UnaryMinus(self, node: UnaryMinus, context: PythonGenerationContext) -> list[str]:
        expr = node.expr.accept(self, context)[0]
        return [f"(- {expr})"]

    def visit_BinaryArithmetic(self, node: BinaryArithmetic, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_Div(self, node: Div, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} / {right})"]

    def visit_Remain(self, node: Remain, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} % {right})"]

    def visit_Mult(self, node: Mult, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} * {right})"]

    def visit_Plus(self, node: Plus, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} + {right})"]

    def visit_Minus(self, node: Minus, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"({left} - {right})"]

    def visit_Concatenation(self, node: Concatenation, context: PythonGenerationContext) -> list[str]:
        left = node.left.accept(self, context)[0]
        right = node.right.accept(self, context)[0]
        return [f"(str({left}) + str({right}))"]

    def visit_InstanceOf(self, node: InstanceOf, context: PythonGenerationContext) -> list[str]:
        instance = node.instance.accept(self, context)
        the_type = node.type.accept(self, context)
        return [f"isinstance({instance}, {the_type})"]

    def visit_Reference(self, node: Reference, context: PythonGenerationContext) -> list[str]:
        return [node.definition.name]

    def visit_Literal(self, node: Literal, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_IntLiteral(self, node: IntLiteral, context: PythonGenerationContext) -> list[str]:
        return [f"{node.value}"]

    def visit_StringLiteral(self, node: StringLiteral, context: PythonGenerationContext) -> list[str]:
        return [f"'{node.value}'"]

    def visit_BoolLiteral(self, node: BoolLiteral, context: PythonGenerationContext) -> list[str]:
        return [f"{'True' if node.value else 'False'}"]

    def visit_RealLiteral(self, node: RealLiteral, context: PythonGenerationContext) -> list[str]:
        return [f"{node.value}"]

    def visit_NullLiteral(self, node: NullLiteral, context: PythonGenerationContext) -> list[str]:
        return [f"None"]

    def visit_EnumLiteral(self, node: EnumLiteral, context: PythonGenerationContext) -> list[str]:
        return [f"{node.enumeration.enum.name}.{node.name}"]

    def visit_SequenceLiteral(self, node: SequenceLiteral, context: PythonGenerationContext) -> list[str]:
        elems = []
        for elem in node.values:
            elems += elem.accept(self, context)
        return [f"[{', '.join(elems)}]"]

    def visit_RangeLiteral(self, node: RangeLiteral, context: PythonGenerationContext) -> list[str]:
        first = node.first.accept(self, context)
        last = node.last.accept(self, context)
        return [f"list(range({first}, 1+{last}))"]

    def visit_Iterator(self, node: Iterator, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_CondLoop(self, node: CondLoop, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_While(self, node: While, context: PythonGenerationContext) -> list[str]:
        cond = node.condition.accept(self, context)[0]
        lines = [f"while {cond}:"]
        lines += indent(node.body.accept(self, context))
        return lines

    def visit_DoWhile(self, node: DoWhile, context: PythonGenerationContext) -> list[str]:
        cond = node.condition.accept(self, context)[0]
        body = node.body.accept(self, context)
        body += [f"if not ({cond}):", "\tbreak"]
        lines = ["while True:"]
        lines += indent(body)
        return lines

    def visit_ConditionalBranch(self, node: ConditionalBranch, context: PythonGenerationContext) -> list[str]:
        pass

    def visit_Block(self, node: Block, context: PythonGenerationContext) -> list[str]:
        statements = []
        for statement in node.statements:
            statements += statement.accept(self, context)

        if len(statements) > 0:
            return statements
        else:
            return ["pass"]

    def visit_Condition(self, node: Condition, context: PythonGenerationContext) -> list[str]:
        cond = node.condition.accept(self, context)[0]
        lines = [f"if {cond}:"]
        then = node.then.accept(self, context)
        lines += indent(then)

        if isinstance(node.elze, Condition):
            elseif = node.elze.accept(self, context)
            elseif[0] = "el" + elseif[0]
            lines += elseif
        elif isinstance(node.elze, Block):
            elze = node.elze.accept(self, context)
            lines.append("else:")
            lines += indent(elze)

        return lines




from besser.BUML.metamodel.action_language.action_language import Type, FunctionType, IntType, StringType, RealType, \
    BoolType, SequenceType, ObjectType, AnyType, OptionalType, EnumType, TypeUnion, Nothing


class UnknownClassifier(ObjectType, EnumType):
    def __init__(self, name: str):
        ObjectType.__init__(self, None)
        EnumType.__init__(self, None)
        self.__name = name

    def name(self) -> str:
        return self.__name


base_classes = {
    'int':          IntType(),
    'float':        RealType(),
    'str':          StringType(),
    'bool':         BoolType(),
    'time':         AnyType(),
    'date':         AnyType(),
    'datetime':     AnyType(),
    'timedelta':    AnyType(),
    'any':          AnyType()
}

def functions_for_sequence_type(sequence_type: SequenceType):
    elem = sequence_type.elementsType
    elem_to_bool = FunctionType([elem],BoolType())
    elem_to_any = FunctionType([elem], AnyType())
    reduce = FunctionType([elem, elem], elem)
    sequence_functions = {
        'size':         FunctionType([],                IntType()),
        'is_empty':     FunctionType([],                BoolType()),
        'add':          FunctionType([elem],  None),
        'remove':       FunctionType([elem],  None),
        'contains':     FunctionType([elem],            BoolType()),
        'filter':       FunctionType([elem_to_bool],    sequence_type),
        'forall':       FunctionType([elem_to_bool],    BoolType()),
        'exists':       FunctionType([elem_to_bool],    BoolType()),
        'one':          FunctionType([elem_to_bool],    BoolType()),
        'is_unique':    FunctionType([elem_to_any],     BoolType()),
        'map':          FunctionType([elem_to_any],     SequenceType(AnyType())),
        'reduce':       FunctionType([reduce, elem],    elem)
    }
    return sequence_functions

def get_typename(t: Type) -> str:
    if t == AnyType():
        return "any"
    if t in base_classes.values():
        for name in base_classes:
            if base_classes[name] == t:
                return name
    if isinstance(t, ObjectType):
        return t.clazz.name
    if isinstance(t, EnumType):
        return t.enum.name
    if isinstance(t, SequenceType):
        return f"[{get_typename(t.elementsType)}]"
    if isinstance(t, OptionalType):
        return f"{get_typename(t.type)}?"
    if isinstance(t, FunctionType):
        param_typenames = list(map(lambda param_type: get_typename(param_type), t.params_type))
        return f"({', '.join(param_typenames)}) -> {get_typename(t.return_type)}"
    if isinstance(t, TypeUnion):
        return f"({get_typename(t.a)}|{get_typename(t.b)})"
    if isinstance(t, Nothing):
        return f"Nothing"
    return f"No Type Infered"



from besser.BUML.metamodel.object import Object, AttributeLink, DataValue

class ObjectBuilder:
    def __init__(self, classifier):
        self.classifier = classifier
        self._name = None
        self._attributes = {}
        self._links = []

    def name(self, name):
        """Set the name of the object."""
        self._name = name
        return self

    def attributes(self, **kwargs):
        """Set attributes of the object."""
        self._attributes.update(kwargs)
        return self

    def link(self, target_obj, end_name:str):
        """Create a link to another object using the association end name."""
        # Find the end and association by end_name
        end_ = None
        for end in self.classifier.all_association_ends():
            if end.name == end_name:
                end_ = end
                break
        if not end_:
            raise ValueError(f"Association end '{end_name}' not found in class '{self.classifier.name}'")

        self._links.append((target_obj, end_))
        return self

    def build(self):
        """Build the Object instance."""
        if not self._name:
            raise ValueError("Object must have a name")

        # Build Object
        obj = Object(name=self._name, classifier=self.classifier)

        # Build AttributeLinks
        for attr_name, value in self._attributes.items():
            # Special handling for "name" attribute to avoid shadowing the object's name property
            # Create AttributeLink directly instead of using setattr
            if attr_name == "name":
                # Find the "name" property in the classifier
                name_property = None
                for attr in self.classifier.attributes | self.classifier.inherited_attributes():
                    if attr.name == "name":
                        name_property = attr
                        break
                
                if name_property:
                    # Create AttributeLink directly for the "name" attribute
                    data_value = DataValue(classifier=name_property.type, value=value)
                    attr_link = AttributeLink(value=data_value, attribute=name_property)
                    obj.add_slot(attr_link)
                else:
                    # If "name" is not a classifier attribute, use setattr (will go through __setattr__)
                    setattr(obj, attr_name, value)
            else:
                setattr(obj, attr_name, value)

        # Build Links
        for target, tgt_end in self._links:
            setattr(obj, tgt_end.name, target)

        return obj



from besser.BUML.metamodel.structural import DomainModel, MethodImplementationType, Method, Class

from besser.BUML.metamodel.action_language.action_language import *
from besser.BUML.notations.action_language.ActionLanguageASTBuilder import parse_bal


def get_model() -> DomainModel:
    from besser.BUML.metamodel.structural import Property, IntegerType, FloatType, DateType, \
        TimeType, StringType, BinaryAssociation, Multiplicity, Parameter
    # Classes
    Book = Class(name="Book")
    Library = Class(name="Library")
    Author = Class(name="Author")

    # Book class attributes and methods
    Book_title: Property = Property(name="title", type=StringType)
    Book_pages: Property = Property(name="pages", type=IntegerType)
    Book_stock: Property = Property(name="stock", type=IntegerType)
    Book_price: Property = Property(name="price", type=FloatType)
    Book_release: Property = Property(name="release", type=DateType)
    Book_time: Property = Property(name="time", type=TimeType)
    Book_m_decrease_stock: Method = Method(name="decrease_stock", parameters=[Parameter(name='qty', type=IntegerType)], implementation_type=MethodImplementationType.NONE)

    Book.attributes={Book_title, Book_release, Book_pages, Book_stock, Book_time, Book_price}
    Book.methods={Book_m_decrease_stock}

    # Library class attributes and methods
    Library_name: Property = Property(name="name", type=StringType)
    Library.attributes={Library_name}

    # Author class attributes and methods
    Author_name: Property = Property(name="name", type=StringType)
    Author.attributes={Author_name}

    # Relationships
    books: BinaryAssociation = BinaryAssociation(
        name="books",
        ends={
            Property(name="library", type=Library, multiplicity=Multiplicity(0, 9999)),
            Property(name="books", type=Book, multiplicity=Multiplicity(0, 9999))
        }
    )
    books_1: BinaryAssociation = BinaryAssociation(
        name="books_1",
        ends={
            Property(name="authors", type=Author, multiplicity=Multiplicity(0, 9999)),
            Property(name="books", type=Book, multiplicity=Multiplicity(1, 9999))
        }
    )

    return DomainModel(
        name="Class_Diagram",
        types={Book, Library, Author},
        associations={books, books_1},
        generalizations={},
        metadata=None
    )

### TODO : Test the typechecker